/**********************************************************************
 *
 * Open Standard Print API (PAPI) - Base High-level Functions
 *
 * Copyright (c) 2002 International Business Machines Corp.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "papi.h"
#include "papi_int.h"

/*---------------------------------------------------------------------
 * Static variables
 *-------------------------------------------------------------------*/
static char msg_buff[100];

/*---------------------------------------------------------------------
 * Service API
 *-------------------------------------------------------------------*/

/**********************************************************************
 *  papiServiceCreate
 *********************************************************************/
papi_status_t papiServiceCreate(
         papi_service_t*         handle,
         char*             service_name,
         char*             user_name,
         char*             password,
         int (*authCB)(papi_service_t svc),
         papi_encryption_t encryption,
         void*                   app_data )
{
	papi_int_service_t* int_handle;

	papi_init();
	papi_dbg_printf(">papiServiceCreate\n");

	*handle = NULL;
	/* Allocate and initialize the space for the handle */
	int_handle = malloc(sizeof(papi_int_service_t));
	if (int_handle == NULL)
		PAPI_RETURN("papiServiceCreate", handle, PAPI_INTERNAL_ERROR);

	memset(int_handle, 0, sizeof(papi_int_service_t));

	/* Set the fields in the handle */
	if (service_name)
		int_handle->service_name = strdup(service_name);
	if (user_name)
		int_handle->user_name = strdup(user_name);
	else
		int_handle->user_name = strdup(ippUser());
	if (password)
		int_handle->password = strdup(password);
	int_handle->authCB = authCB;
	int_handle->encryption = encryption;
	int_handle->app_data = app_data;

	papi_dbg_printf("int_handle=0x%x\n", int_handle);
	*handle = (papi_service_t*)int_handle;

	papi_dbg_printf("<papiServiceCreate returning OK\n");
	return(PAPI_OK);
}

/**********************************************************************
 *  papiServiceDestroy
 *********************************************************************/
void papiServiceDestroy(
         papi_service_t handle )
{
	if (handle)
	{
		papi_clear_prev_request(INT_HANDLE(handle));
		if (INT_HANDLE(handle)->service_name)
			free(INT_HANDLE(handle)->service_name);
		if (INT_HANDLE(handle)->user_name)
			free(INT_HANDLE(handle)->user_name);
		if (INT_HANDLE(handle)->password)
			free(INT_HANDLE(handle)->password);
		free(INT_HANDLE(handle));
	}	
}

/**********************************************************************
 *  papiServiceSetUserName
 *********************************************************************/
papi_status_t papiServiceSetUserName(
         papi_service_t handle,
         char* user_name )
{
	if (handle == NULL)
		PAPI_RETURN("papiServiceSetUserName", handle, PAPI_BAD_REQUEST);

	if (INT_HANDLE(handle)->user_name)
		free(INT_HANDLE(handle)->user_name);

	INT_HANDLE(handle)->user_name = NULL;
	if (user_name)
		INT_HANDLE(handle)->user_name = strdup(user_name);
	
	return(PAPI_OK);
}

/**********************************************************************
 *  papiServiceSetPassword
 *********************************************************************/
papi_status_t papiServiceSetPassword(
         papi_service_t handle,
         char* password )
{
	if (handle == NULL)
		PAPI_RETURN("papiServiceSetPassword", handle, PAPI_BAD_REQUEST);

	if (INT_HANDLE(handle)->password)
		free(INT_HANDLE(handle)->password);

	INT_HANDLE(handle)->password = NULL;
	if (password)
		INT_HANDLE(handle)->password = strdup(password);
	
	return(PAPI_OK);
}

/**********************************************************************
 *  papiServiceSetEncryption
 *********************************************************************/
papi_status_t papiServiceSetEncryption(
         papi_service_t handle,
         papi_encryption_t encryption )
{
	if (handle == NULL)
		PAPI_RETURN("papiServiceSetEncryption", handle, PAPI_BAD_REQUEST);

	INT_HANDLE(handle)->encryption = encryption;
	
	return(PAPI_OK);
}

/**********************************************************************
 *  papiServiceSetAuthCB
 *********************************************************************/
papi_status_t papiServiceSetAuthCB(
         papi_service_t handle,
         int (*authCB)(papi_service_t svc) )
{
	if (handle == NULL)
		PAPI_RETURN("papiServiceSetAuthCB", handle, PAPI_BAD_REQUEST);

	INT_HANDLE(handle)->authCB = authCB;
	
	return(PAPI_OK);
}

/**********************************************************************
 *  papiServiceSetAddData
 *********************************************************************/
papi_status_t papiServiceSetAppData(
         papi_service_t handle,
         void*    app_data )
{
	if (handle == NULL)
		PAPI_RETURN("papiServiceSetAppData", handle, PAPI_BAD_REQUEST);

	if (INT_HANDLE(handle)->app_data)
		return(PAPI_BAD_REQUEST);

	INT_HANDLE(handle)->app_data = app_data;
	
	return(PAPI_OK);
}

/**********************************************************************
 *  papiServiceGetServiceName
 *********************************************************************/
char* papiServiceGetServiceName(
         papi_service_t handle )
{
	if (handle == NULL)
		return(NULL);
	return(INT_HANDLE(handle)->service_name);
}

/**********************************************************************
 *  papiServiceGetUserName
 *********************************************************************/
char* papiServiceGetUserName(
         papi_service_t handle )
{
	if (handle == NULL)
		return(NULL);
	return(INT_HANDLE(handle)->user_name);
}

/**********************************************************************
 *  papiServiceGetPassword
 *********************************************************************/
char* papiServiceGetPassword(
         papi_service_t handle )
{
	if (handle == NULL)
		return(NULL);
	return(INT_HANDLE(handle)->password);
}

/**********************************************************************
 *  papiServiceGetEncryption
 *********************************************************************/
papi_encryption_t papiServiceGetEncryption(
         papi_service_t handle )
{
	if (handle == NULL)
		return(PAPI_ENCRYPT_NEVER);
	return(INT_HANDLE(handle)->encryption);
}

/**********************************************************************
 *  papiServiceGetAppData
 *********************************************************************/
void* papiServiceGetAppData(
         papi_service_t handle )
{
	if (handle == NULL)
		return(NULL);
	return(INT_HANDLE(handle)->app_data);
}

papi_attribute_t **
papiServiceGetAttributeList(papi_service_t handle)
{
	return((papi_attribute_t **)NULL);
}

/**********************************************************************
 *  papiServiceGetStatusMessage
 *********************************************************************/
char* papiServiceGetStatusMessage(
         papi_service_t handle )
{
	char* msg = "";
	ipp_attribute_t* attr;

	if (handle)
	{
		if (INT_HANDLE(handle)->response)
		{
			attr = ippFindAttribute(INT_HANDLE(handle)->response, "status-message", IPP_TAG_NAME);
			if (attr == NULL)
				msg = papiStatusString(INT_HANDLE(handle)->last_error);
			else
				msg = attr->values[0].string.text;
		}
		else
		{
			msg = papiStatusString(INT_HANDLE(handle)->last_error);
		}
	}
	return(msg);
}

/*---------------------------------------------------------------------
 * Miscellaneous API
 *-------------------------------------------------------------------*/

/**********************************************************************
 *  papiStatusString
 *********************************************************************/
char* papiStatusString(
          papi_status_t status )
{
	char* msg = NULL;

	switch(status)
	{
	case PAPI_OK                             : msg = "ok"; break;
	case PAPI_OK_SUBST                       : msg = "ok-subst"; break;
	case PAPI_OK_CONFLICT                    : msg = "ok-conflict"; break;
	case PAPI_OK_IGNORED_SUBSCRIPTIONS       : msg = "ok-ignored-subscriptions"; break;
	case PAPI_OK_IGNORED_NOTIFICATIONS       : msg = "ok-ignored-notifications"; break;
	case PAPI_OK_TOO_MANY_EVENTS             : msg = "ok-too-many-events"; break;
	case PAPI_OK_BUT_CANCEL_SUBSCRIPTION     : msg = "ok-but-cancel-subscription"; break;
	case PAPI_REDIRECTION_OTHER_SITE         : msg = "redirection-other-site"; break;
	case PAPI_BAD_REQUEST                    : msg = "bad-request"; break;
	case PAPI_FORBIDDEN                      : msg = "forbidden"; break;
	case PAPI_NOT_AUTHENTICATED              : msg = "not-authenticated"; break;
	case PAPI_NOT_AUTHORIZED                 : msg = "not-authorized"; break;
	case PAPI_NOT_POSSIBLE                   : msg = "not-possible"; break;
	case PAPI_TIMEOUT                        : msg = "timeout"; break;
	case PAPI_NOT_FOUND                      : msg = "not-found"; break;
	case PAPI_GONE                           : msg = "gone"; break;
	case PAPI_REQUEST_ENTITY                 : msg = "request-entity"; break;
	case PAPI_REQUEST_VALUE                  : msg = "request-value"; break;
	case PAPI_DOCUMENT_FORMAT                : msg = "document-format"; break;
	case PAPI_ATTRIBUTES                     : msg = "attributes"; break;
	case PAPI_URI_SCHEME                     : msg = "uri-scheme"; break;
	case PAPI_CHARSET                        : msg = "charset"; break;
	case PAPI_CONFLICT                       : msg = "conflict"; break;
	case PAPI_COMPRESSION_NOT_SUPPORTED      : msg = "compression-not-supported"; break;
	case PAPI_COMPRESSION_ERROR              : msg = "compression-error"; break;
	case PAPI_DOCUMENT_FORMAT_ERROR          : msg = "document-format-error"; break;
	case PAPI_DOCUMENT_ACCESS_ERROR          : msg = "document-access-error"; break;
	case PAPI_ATTRIBUTES_NOT_SETTABLE        : msg = "attributes-not-settable"; break;
	case PAPI_IGNORED_ALL_SUBSCRIPTIONS      : msg = "ignored-all-subscriptions"; break;
	case PAPI_TOO_MANY_SUBSCRIPTIONS         : msg = "too-many-subscriptions"; break;
	case PAPI_IGNORED_ALL_NOTIFICATIONS      : msg = "ignored-all-notifications"; break;
	case PAPI_PRINT_SUPPORT_FILE_NOT_FOUND   : msg = "print-support-file-not-found"; break;
	case PAPI_INTERNAL_ERROR                 : msg = "internal-error"; break;
	case PAPI_OPERATION_NOT_SUPPORTED        : msg = "operation-not-supported"; break;
	case PAPI_SERVICE_UNAVAILABLE            : msg = "service-unavailable"; break;
	case PAPI_VERSION_NOT_SUPPORTED          : msg = "version-not-supported"; break;
	case PAPI_DEVICE_ERROR                   : msg = "device-error"; break;
	case PAPI_TEMPORARY_ERROR                : msg = "temporary-error"; break;
	case PAPI_NOT_ACCEPTING                  : msg = "not-accepting"; break;
	case PAPI_PRINTER_BUSY                   : msg = "printer-busy"; break;
	case PAPI_ERROR_JOB_CANCELLED            : msg = "error-job-cancelled"; break;
	case PAPI_MULTIPLE_JOBS_NOT_SUPPORTED    : msg = "multiple-jobs-not-supported"; break;
	case PAPI_PRINTER_IS_DEACTIVATED         : msg = "printer-is-deactivated"; break;
	default:
		sprintf(msg_buff, "unknown error: %d", status);
		msg = msg_buff;
		break;
	}

	return(msg);
}
