/**********************************************************************
 *
 * Open Standard Print API (PAPI) - Base High-level Functions
 *
 * Copyright (c) 2002 International Business Machines Corp.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "papi.h"
#include "papi_int.h"

/*---------------------------------------------------------------------
 * Internal functions
 *-------------------------------------------------------------------*/

/**********************************************************************
 *  papi_add_cups_attr
 *********************************************************************/
ipp_attribute_t* papi_add_cups_attr(ipp_t* req, int num_values)
{
	ipp_attribute_t* attr;  /* New attribute */

	if (req == NULL || num_values < 0)
		return (NULL);

	attr = calloc(sizeof(ipp_attribute_t) +
				(num_values - 1) * sizeof(ipp_value_t), 1);

	attr->num_values = num_values;

	if (attr == NULL)
		return (NULL);

	if (req->last == NULL)
		req->attrs = attr;
	else
		req->last->next = attr;

	req->last = attr;

	return (attr);
}

/**********************************************************************
 *  papi_attr_cups2papi
 *********************************************************************/
papi_status_t papi_attr_cups2papi(ipp_attribute_t* cups_attrs,
                           papi_attribute_t*** p_papi_attrs)
{
	int i;
	int k;
	int attr_count = 0;
	ipp_attribute_t* cups_attr = NULL;
	ipp_value_t* cups_val = NULL;
	papi_dbg_printf(">papi_attr_cups2papi\n");

	/* Allocate the attribute list */
	for(cups_attr=cups_attrs;
        (cups_attr != NULL) &&
          !((cups_attr->name == NULL) && (cups_attr->group_tag == IPP_TAG_ZERO));
        cups_attr=cups_attr->next)
		attr_count++;
	papi_dbg_printf("%d attributes to be converted\n", attr_count);
	*p_papi_attrs = malloc(sizeof(papi_attribute_t*) * (attr_count+1));
	if (*p_papi_attrs == NULL)
		PAPI_RETURN("papi_attr_cups2papi", NULL, PAPI_INTERNAL_ERROR);
	(*p_papi_attrs)[attr_count] = NULL;

	/* Convert each CUPS attribute */	
	for(i=0, cups_attr=cups_attrs;
	    (cups_attr != NULL) &&
          !((cups_attr->name == NULL) && (cups_attr->group_tag == IPP_TAG_ZERO));
		i++, cups_attr=cups_attr->next)
	{
		/* Add the attribute */
		(*p_papi_attrs)[i] = malloc(sizeof(papi_attribute_t));
		if ((*p_papi_attrs)[i] == NULL)
			PAPI_RETURN("papi_attr_cups2papi", NULL, PAPI_INTERNAL_ERROR);
		(*p_papi_attrs)[i]->name = strdup(cups_attr->name);
		papi_dbg_printf("processing %s...\n", (*p_papi_attrs)[i]->name);

		/* Store the PAPI type */
		switch(cups_attr->value_tag)
		{
		case IPP_TAG_CHARSET:
		case IPP_TAG_KEYWORD:
		case IPP_TAG_LANGUAGE:
		case IPP_TAG_MEMBERNAME:
		case IPP_TAG_MIMETYPE:
		case IPP_TAG_NAME:
		case IPP_TAG_NAMELANG:
		case IPP_TAG_NOVALUE:
		case IPP_TAG_STRING:
		case IPP_TAG_TEXT:
		case IPP_TAG_TEXTLANG:
		case IPP_TAG_URI:
		case IPP_TAG_URISCHEME:
			(*p_papi_attrs)[i]->type = PAPI_STRING;
  			break;

		case IPP_TAG_ENUM:
		case IPP_TAG_INTEGER:
			(*p_papi_attrs)[i]->type = PAPI_INTEGER;
			break;

		case IPP_TAG_BOOLEAN:
			(*p_papi_attrs)[i]->type = PAPI_BOOLEAN;
			break;

		case IPP_TAG_DATE:
			(*p_papi_attrs)[i]->type = PAPI_DATETIME;
			break;

		case IPP_TAG_RANGE:
			(*p_papi_attrs)[i]->type = PAPI_RANGE;
			break;

		case IPP_TAG_RESOLUTION:
			(*p_papi_attrs)[i]->type = PAPI_RESOLUTION;
			break;
			
		default:
			papi_dbg_printf("unrecognized tag: 0x%x\n", cups_attr->value_tag);
			PAPI_RETURN("papi_attr_cups2papi", NULL, PAPI_INTERNAL_ERROR);
		}

		/* Process all the values according to the PAPI type */
		papi_dbg_printf("num_values=%d\n", cups_attr->num_values);
		(*p_papi_attrs)[i]->values = malloc( (cups_attr->num_values+1) * sizeof(papi_attribute_value_t*) );
		if ((*p_papi_attrs)[i]->values == NULL)
			PAPI_RETURN("papi_attr_cups2papi", NULL, PAPI_INTERNAL_ERROR);
		(*p_papi_attrs)[i]->values[cups_attr->num_values] = NULL; /* mark end of value list */
		for(k=0; k<cups_attr->num_values; k++)
		{
			(*p_papi_attrs)[i]->values[k] = malloc(sizeof(papi_attribute_value_t));
			if ((*p_papi_attrs)[i]->values[k] == NULL)
				PAPI_RETURN("papi_attr_cups2papi", NULL, PAPI_INTERNAL_ERROR);

 			switch((*p_papi_attrs)[i]->type)
 			{
 			case PAPI_STRING:
 				if (cups_attr->value_tag == IPP_TAG_NOVALUE)
	  	  			(*p_papi_attrs)[i]->values[k]->string = strdup("");
				else
	  	  			(*p_papi_attrs)[i]->values[k]->string = strdup(cups_attr->values[k].string.text);
	  	  		break;

 			case PAPI_BOOLEAN:
  		  		(*p_papi_attrs)[i]->values[k]->boolean = cups_attr->values[k].boolean;
	  	  		break;

 			case PAPI_INTEGER:
  		  		(*p_papi_attrs)[i]->values[k]->integer = cups_attr->values[k].integer;
	  	  		break;

 			case PAPI_DATETIME:
  		  		(*p_papi_attrs)[i]->values[k]->datetime = ippDateToTime(cups_attr->values[k].date);
 				break;

 			case PAPI_RANGE:
  		  		(*p_papi_attrs)[i]->values[k]->range.lower = cups_attr->values[k].range.lower;
  		  		(*p_papi_attrs)[i]->values[k]->range.upper = cups_attr->values[k].range.upper;
 				break;

 			case PAPI_RESOLUTION:
  		  		(*p_papi_attrs)[i]->values[k]->resolution.units = (papi_resolution_unit_t)cups_attr->values[k].resolution.units;
  		  		(*p_papi_attrs)[i]->values[k]->resolution.xres = cups_attr->values[k].resolution.xres;
  		  		(*p_papi_attrs)[i]->values[k]->resolution.yres = cups_attr->values[k].resolution.yres;
 				break;

 			default:
				PAPI_RETURN("papi_attr_cups2papi", NULL, PAPI_INTERNAL_ERROR);
 			}
 		} /* end loop throug CUPS values */

	} /* end loop through CUPS attributes */

	PAPI_RETURN("papi_attr_cups2papi", NULL, PAPI_OK);
}
typedef struct
{
	char* 		attr;
	ipp_tag_t	tag;
	ipp_tag_t	group;
} attr_value_tag_map_t;

/* Table mapping names to IPP tags */
static attr_value_tag_map_t ATTR_VALUE_TAG_MAP[] =
{
	{ "attributes-charset",				IPP_TAG_CHARSET,  IPP_TAG_OPERATION },
	{ "attributes-natural-language",	IPP_TAG_LANGUAGE, IPP_TAG_OPERATION },
	{ "charset-configured",				IPP_TAG_CHARSET,  IPP_TAG_OPERATION },
	{ "charset-supported",				IPP_TAG_CHARSET,  IPP_TAG_OPERATION },
	{ "color-supported",				IPP_TAG_BOOLEAN,  IPP_TAG_OPERATION },
	{ "compression-supported",			IPP_TAG_KEYWORD,  IPP_TAG_OPERATION },
	{ "copies",							IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "document-format",				IPP_TAG_MIMETYPE, IPP_TAG_OPERATION },
	{ "document-format-default",		IPP_TAG_MIMETYPE, IPP_TAG_OPERATION },
	{ "document-format-supported",		IPP_TAG_MIMETYPE, IPP_TAG_OPERATION },
	{ "finishings",						IPP_TAG_ENUM,     IPP_TAG_JOB },
	{ "generated-natural-language-supported",IPP_TAG_LANGUAGE, IPP_TAG_OPERATION },
	{ "job-hold-until",					IPP_TAG_KEYWORD,  IPP_TAG_JOB },
	{ "job-id",							IPP_TAG_INTEGER,  IPP_TAG_OPERATION },
	{ "job-impressions",				IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "job-impressions-completed",		IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "job-impressions-supported",		IPP_TAG_RANGE,    IPP_TAG_OPERATION },
	{ "job-k-octets",					IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "job-k-octets-processed",			IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "job-k-octets-supported",			IPP_TAG_RANGE,    IPP_TAG_OPERATION },
	{ "job-media-sheets",				IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "job-media-sheets-completed",		IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "job-media-sheets-supported",		IPP_TAG_RANGE,    IPP_TAG_OPERATION },
	{ "job-message-from-operator",		IPP_TAG_STRING,   IPP_TAG_JOB },
	{ "job-more-info",					IPP_TAG_URI,      IPP_TAG_JOB },
	{ "job-name",						IPP_TAG_NAME,     IPP_TAG_OPERATION },
	{ "job-originating-user-name",		IPP_TAG_NAME,     IPP_TAG_OPERATION },
	{ "job-printer-uri",				IPP_TAG_URI,      IPP_TAG_OPERATION },
	{ "job-priority",					IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "job-sheets",						IPP_TAG_KEYWORD,  IPP_TAG_JOB },
	{ "job-state",						IPP_TAG_ENUM,     IPP_TAG_OPERATION },
	{ "job-state-message",				IPP_TAG_STRING,   IPP_TAG_JOB },
	{ "job-state-reasons",				IPP_TAG_KEYWORD,  IPP_TAG_OPERATION },
	{ "job-uri",						IPP_TAG_URI,      IPP_TAG_OPERATION },
	{ "media", 							IPP_TAG_KEYWORD,  IPP_TAG_JOB },
	{ "message", 						IPP_TAG_STRING,   IPP_TAG_OPERATION },
	{ "multiple-document-handling",		IPP_TAG_KEYWORD,  IPP_TAG_JOB },
	{ "multiple-operation-time-out",	IPP_TAG_INTEGER,  IPP_TAG_OPERATION },
	{ "natural-language-configured",	IPP_TAG_LANGUAGE, IPP_TAG_OPERATION },
	{ "number-of-documents", 			IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "number-of-intervening-jobs",		IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "number-up",						IPP_TAG_INTEGER,  IPP_TAG_JOB },
	{ "operations-supported",			IPP_TAG_ENUM,     IPP_TAG_OPERATION },
	{ "orientation-requested",			IPP_TAG_ENUM,     IPP_TAG_JOB },
	{ "output-device-assigned",			IPP_TAG_NAME,     IPP_TAG_JOB },
	{ "page-ranges",					IPP_TAG_RANGE,    IPP_TAG_JOB },
	{ "pdl-override-supported",			IPP_TAG_KEYWORD,  IPP_TAG_OPERATION },
	{ "print-quality",					IPP_TAG_ENUM,     IPP_TAG_JOB },
	{ "printer-current-time",			IPP_TAG_DATE,     IPP_TAG_OPERATION },
	{ "printer-driver-installer",		IPP_TAG_URI,      IPP_TAG_OPERATION },
	{ "printer-info",					IPP_TAG_STRING,   IPP_TAG_OPERATION },
	{ "printer-is-accepting-jobs",		IPP_TAG_BOOLEAN,  IPP_TAG_OPERATION },
	{ "printer-location",				IPP_TAG_STRING,   IPP_TAG_OPERATION },
	{ "printer-make-and-model",			IPP_TAG_STRING,   IPP_TAG_OPERATION },
	{ "printer-message-from-operator",	IPP_TAG_STRING,   IPP_TAG_OPERATION },
	{ "printer-more-info",				IPP_TAG_URI,      IPP_TAG_OPERATION },
	{ "printer-more-info-manufacturer",	IPP_TAG_URI,      IPP_TAG_OPERATION },
	{ "printer-name",					IPP_TAG_NAME,     IPP_TAG_OPERATION },
	{ "printer-resolution",				IPP_TAG_RESOLUTION, IPP_TAG_JOB },
	{ "printer-state",					IPP_TAG_ENUM,     IPP_TAG_OPERATION },
	{ "printer-state-message",			IPP_TAG_STRING,   IPP_TAG_OPERATION },
	{ "printer-state-reasons",			IPP_TAG_KEYWORD,  IPP_TAG_OPERATION },
	{ "printer-up-time",				IPP_TAG_INTEGER,  IPP_TAG_OPERATION },
	{ "printer-uri",					IPP_TAG_URI,      IPP_TAG_OPERATION },
	{ "printer-uri-supported",			IPP_TAG_URI,      IPP_TAG_OPERATION },
	{ "queued-job-count",				IPP_TAG_INTEGER,  IPP_TAG_OPERATION },
	{ "reference-uri-schemes-supported",IPP_TAG_URISCHEME,IPP_TAG_OPERATION },
	{ "requesting-user-name",			IPP_TAG_NAME,     IPP_TAG_OPERATION },
	{ "sides",							IPP_TAG_KEYWORD,  IPP_TAG_JOB },
	{ "time-at-completed",				IPP_TAG_INTEGER,  IPP_TAG_OPERATION },
	{ "time-at-creation",				IPP_TAG_INTEGER,  IPP_TAG_OPERATION },
	{ "time-at-processing",				IPP_TAG_INTEGER,  IPP_TAG_OPERATION },
	{ "uri-security-supported",			IPP_TAG_KEYWORD,  IPP_TAG_OPERATION },
	/* End of table */
	{ NULL, IPP_TAG_STRING, IPP_TAG_OPERATION }
};


/**********************************************************************
 *  papi_attr_get_tag
 *********************************************************************/
ipp_tag_t papi_attr_get_tag(char* attr_name, ipp_tag_t dflt_tag)
{
	int i;
		
	for(i=0; ATTR_VALUE_TAG_MAP[i].attr != NULL; i++)	
	{
		if ( !strcmp(ATTR_VALUE_TAG_MAP[i].attr, attr_name) )
			break;
	}

	if (ATTR_VALUE_TAG_MAP[i].attr != NULL)
		return( ATTR_VALUE_TAG_MAP[i].tag );
	else
		return( dflt_tag );
}

/**********************************************************************
 *  papi_attr_get_tag_and_group
 *********************************************************************/
void papi_attr_get_tag_and_group(char* attr_name, ipp_tag_t* tag, ipp_tag_t* group)
{
	int i;
		
	for(i=0; ATTR_VALUE_TAG_MAP[i].attr != NULL; i++)	
	{
		if ( !strcmp(ATTR_VALUE_TAG_MAP[i].attr, attr_name) )
			break;
	}

	if (ATTR_VALUE_TAG_MAP[i].attr != NULL)
	{
		*tag = ATTR_VALUE_TAG_MAP[i].tag;
		*group = ATTR_VALUE_TAG_MAP[i].group;
	}
}

/**********************************************************************
 * papi_attr_papi2cups
 *********************************************************************/
papi_status_t papi_attr_papi2cups(papi_attribute_t** papi_attrs,
						   ipp_t* request,
						   char* char_set,
						   ipp_tag_t attr_group)
{
	papi_status_t sc;
	int i;
	int k;
	ipp_tag_t tag;
	ipp_tag_t group;
	ipp_attribute_t* cups_attr;
	papi_attribute_t* papi_attr;
	int value_count = 0;
	char** values_char;
	char*  values_bool;
	int*   values_int;
	int*   values_int2;
	ipp_uchar_t* ipp_date;

	papi_dbg_printf(">papi_attr_papi2cups\n");

	if (papi_attrs == NULL)
	{
		papi_dbg_printf("<papi_attr_papi2cups returning OK (no attrs)\n");
		return(PAPI_OK);
	}

	/* Loop through all the PAPI attributes and convert them */
	for(i=0; papi_attrs[i] != NULL; i++)
	{
		papi_dbg_printf("processing %s...\n", papi_attrs[i]->name);

		/* Get the tag and group */
		tag = IPP_TAG_STRING; /* default */
		group = IPP_TAG_JOB; /* default */
		papi_attr_get_tag_and_group(papi_attrs[i]->name, &tag, &group);

		/* See if this is in the attr group currently being requested */
		if ((attr_group != IPP_TAG_ZERO) && (group != attr_group))
		{
			papi_dbg_printf("skip it\n");
			continue;
		}

		/* Get the number of values */
		value_count = 0;
		if (papi_attrs[i]->values != NULL)
		{
			for(k=0; papi_attrs[i]->values[k] != NULL; k++)
				value_count++;
		}

		/* Add the attribute and values */
		switch(papi_attrs[i]->type)
		{
		case PAPI_STRING:
			values_char = malloc( sizeof(char*) * value_count );
			if (values_char == NULL)
				PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_INTERNAL_ERROR);
			if (papi_attrs[i]->values != NULL)
			{
				for(k=0; papi_attrs[i]->values[k] != NULL; k++)
	 				values_char[k] = papi_attrs[i]->values[k]->string;
			}

			cups_attr = ippAddStrings(request,
							group,
							tag,
							papi_attrs[i]->name,
							value_count,
							char_set,
							values_char);
			free(values_char);
			if (cups_attr == NULL)
			{
				sc = (papi_status_t)ippLastError();
				PAPI_RETURN("papi_attr_papi2cups", NULL, sc);
			}
 	  		break;

		case PAPI_BOOLEAN:
			values_bool = malloc( sizeof(char) * value_count );
			if (values_bool == NULL)
				PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_INTERNAL_ERROR);
			if (papi_attrs[i]->values != NULL)
			{
				for(k=0; papi_attrs[i]->values[k] != NULL; k++)
	 				values_bool[k] = papi_attrs[i]->values[k]->boolean;
			}

			cups_attr = ippAddBooleans(request,
							group,
							papi_attrs[i]->name,
							value_count,
							values_bool);
			free(values_bool);
			if (cups_attr == NULL)
			{
				sc = (papi_status_t)ippLastError();
				PAPI_RETURN("papi_attr_papi2cups", NULL, sc);
			}
 	  		break;

		case PAPI_INTEGER:
			values_int = malloc( sizeof(int) * value_count );
			if (values_int == NULL)
				PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_INTERNAL_ERROR);
			if (papi_attrs[i]->values != NULL)
			{
				for(k=0; papi_attrs[i]->values[k] != NULL; k++)
	 				values_int[k] = papi_attrs[i]->values[k]->integer;
			}	

			cups_attr = ippAddIntegers(request,
							group,
							tag,
							papi_attrs[i]->name,
							value_count,
							values_int);
			free(values_int);
			if (cups_attr == NULL)
			{
				sc = (papi_status_t)ippLastError();
				PAPI_RETURN("papi_attr_papi2cups", NULL, sc);
			}
 	  		break;

		case PAPI_RANGE:
			values_int = malloc( sizeof(int) * value_count );
			if (values_int == NULL)
				PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_INTERNAL_ERROR);
			values_int2 = malloc( sizeof(int) * value_count );
			if (values_int2 == NULL)
				PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_INTERNAL_ERROR);
			if (papi_attrs[i]->values != NULL)
			{
				for(k=0; papi_attrs[i]->values[k] != NULL; k++)
				{
	 				values_int[k] = papi_attrs[i]->values[k]->range.lower;
	 				values_int2[k] = papi_attrs[i]->values[k]->range.upper;
	 			}
			}

			cups_attr = ippAddRanges(request,
							group,
							papi_attrs[i]->name,
							value_count,
							values_int,
							values_int2);
			free(values_int);
			free(values_int2);
			if (cups_attr == NULL)
			{
				sc = (papi_status_t)ippLastError();
				PAPI_RETURN("papi_attr_papi2cups", NULL, sc);
			}
 	  		break;

		case PAPI_RESOLUTION:
			values_int = malloc( sizeof(int) * value_count );
			if (values_int == NULL)
				PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_INTERNAL_ERROR);
			values_int2 = malloc( sizeof(int) * value_count );
			if (values_int2 == NULL)
				PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_INTERNAL_ERROR);
			if (papi_attrs[i]->values != NULL)
			{
				for(k=0; papi_attrs[i]->values[k] != NULL; k++)
				{
	 				values_int[k] = papi_attrs[i]->values[k]->resolution.xres;
	 				values_int2[k] = papi_attrs[i]->values[k]->resolution.yres;
	 			}
			}

			cups_attr = ippAddResolutions(request,
							group,
							papi_attrs[i]->name,
							value_count,
							(ipp_res_t)papi_attrs[i]->values[0]->resolution.units,
							values_int,
							values_int2);
			free(values_int);
			free(values_int2);
			if (cups_attr == NULL)
			{
				sc = (papi_status_t)ippLastError();
				PAPI_RETURN("papi_attr_papi2cups", NULL, sc);
			}
 	  		break;

		case PAPI_DATETIME:
			ipp_date = ippTimeToDate(papi_attrs[i]->values[0]->datetime);
			cups_attr = ippAddDate(request,
								group,
								papi_attrs[i]->name,
								ipp_date);
			if (cups_attr == NULL)
			{
				sc = (papi_status_t)ippLastError();
				PAPI_RETURN("papi_attr_papi2cups", NULL, sc);
			}
 			break;

 		default:
			PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_INTERNAL_ERROR);
 		}

	}

	PAPI_RETURN("papi_attr_papi2cups", NULL, PAPI_OK);
}

/**********************************************************************
 * papi_values_free
 *********************************************************************/
void papi_values_free(papi_attribute_t* attr)
{
	int i;

	if (attr)
	{
		if (attr->values)
		{
			for(i=0; attr->values[i] != NULL; i++)
			{
				if ((attr->type == PAPI_STRING) && (attr->values[i]->string))
					free(attr->values[i]->string);
				free(attr->values[i]);				
			}
			free(attr->values);
			attr->values = NULL;
		}
	}
}

/**********************************************************************
 * papi_attr_free
 *********************************************************************/
void papi_attr_free(papi_attribute_t* attr)
{
	if (attr)
	{
		if (attr->name)
			free(attr->name);
		papi_values_free(attr);
		free(attr);
	}
}

/**********************************************************************
 *  papi_skip_whitespace
 *********************************************************************/
void papi_skip_whitespace(char** str)
{
	while((**str != '\0') && (**str != ';'))
	{
		if ((**str != ' ') &&
		    (**str != '\n') &&
		    (**str != '\t') )
			break;
		(*str)++;
	}
}

/**********************************************************************
 *  papi_get_octal_value
 *********************************************************************/
unsigned char papi_get_octal_value(char** str)
{
	unsigned char ret = 0;
	char* ch = *str;
	char conv[2] = { '\0', '\0' };
	int d0;
	int d1;
	int d2;

	if ( (isdigit(ch[0]) && (ch[0] != '8') && (ch[0] != '9')) &&
         (isdigit(ch[1]) && (ch[1] != '8') && (ch[1] != '9')) &&
         (isdigit(ch[2]) && (ch[2] != '8') && (ch[2] != '9')) )
    {
    	conv[0] = ch[0];
    	d0 = atoi(conv);
    	conv[0] = ch[1];
    	d1 = atoi(conv);
    	conv[0] = ch[2];
    	d2 = atoi(conv);

    	ret = (d0*64) + (d1*8) + d2;
    	*str = *str + 3;
    }

	return(ret);
}

/**********************************************************************
 *  papi_get_quoted_value_str
 *********************************************************************/
int papi_get_quoted_value_str(char quote, char** str, char* value_str)
{
	int rc = 1;
	int i;
	unsigned char octal;

	/* Parse single-quoted value string */
	(*str)++;
	for(i=0;
	    ( (**str != quote) &&
		  (**str != '\0') &&
		  (i < IPP_MAX_NAME-1) );
		i++)
	{	
		if (**str == '\\')
		{
			(*str)++;
			octal = papi_get_octal_value(str);
			if (octal == 0)
				i--; /* not an octal so take literal */
			else
				value_str[i] = octal; /* store the binary */
		}
		else
		{
			value_str[i] = **str;			
			(*str)++;
		}
	}
	if (**str != quote)
		rc = 0;
	else
		(*str)++;

	if (i < IPP_MAX_NAME)
		value_str[i] = '\0';
	else
		return(0);

	return(rc);
}

/**********************************************************************
 *  papi_get_value_str
 *********************************************************************/
int papi_get_value_str(char** str, char* value_str)
{
	int rc = 1;
	int i;

	papi_skip_whitespace(str);

	if (**str == '\'')
	{
		/* Parse single-quoted value string */
		rc = papi_get_quoted_value_str('\'', str, value_str);
	}
	else if (**str == '"')
	{
		/* Parse double-quoted value string */
		rc = papi_get_quoted_value_str('"', str, value_str);
	}
	else
	{
		/* Parse non-quoted value string */
		for(i=0;
	    ( (**str != ' ') &&
		  (**str != ',') &&
		  (**str != '\n') &&
		  (**str != '\0') &&
		  (i < IPP_MAX_NAME-1) );
		i++)
		{	
			value_str[i] = **str;			
			(*str)++;
		}

		if (i < IPP_MAX_NAME)
			value_str[i] = '\0';
		else
			return(0);
	}

	if (**str == ',')
		(*str)++;
    else if ((**str != ' ')&&(**str != '\n')&&(**str != '\0'))
		rc = 0;

	return(rc);
}

/**********************************************************************
 *  papi_parse_attr_get_value_count
 *********************************************************************/
int papi_parse_attr_get_value_count(char* str)
{
	int count = 0;
	char* attr_str = str;
	char value_str[IPP_MAX_NAME];

	/* Find the start of the values */
	while((*attr_str != '=') && (*attr_str != ';') && (*attr_str != '\0'))
		attr_str++;

	if (*attr_str == '=')
	{
		attr_str++;		

		/* Parse through them, counting each value */
		while((*attr_str != '=') && (*attr_str != ';') && (*attr_str != '\0'))
		{
			if (!papi_get_value_str(&attr_str, value_str))
				return(0);
			count++;
			papi_skip_whitespace(&attr_str);
		}
	}

	return(count);
}

/**********************************************************************
 *  papi_get_str_digits_value
 *********************************************************************/
int papi_get_str_digits_value(char* buffer, int num_digits)
{
	int ret = -1;
	char save_ch;

	save_ch = buffer[num_digits];
	buffer[num_digits] = '\0';
	ret = atoi(buffer);
	buffer[num_digits] = save_ch;

	return(ret);	
}

/**********************************************************************
 *  papi_parse_value
 *********************************************************************/
int papi_parse_value(int k, char** attr_str, int add_flags,
            papi_attribute_t* attr, papi_attribute_t*** attrs)
{
	papi_status_t sc;
	char value_str[IPP_MAX_NAME];
	ipp_value_t* temp_values;
	int int_val;
	char* low;
	char* high;
	char* res;
	int i;
	struct tm tm;
	struct tm* p_tm;
	time_t ltime;
	int len;
	int xres;
	int yres;
	papi_resolution_unit_t units;

	/* Skip over preceding whitespace */
	papi_skip_whitespace(attr_str);
	if ((**attr_str == '\0') || (**attr_str == ';') || (**attr_str == '='))
	{
		papi_dbg_printf("Syntax error in attributes at line %d\n", __LINE__);
		papi_dbg_printf("*attr_str: \"%s\"\n", *attr_str);
		return(0);
	}

	/* Get the value string */
	if (!papi_get_value_str(attr_str, value_str))
	{
		papi_dbg_printf("Syntax error in attributes at line %d\n", __LINE__);
		return(0);
	}
	
	/* Convert/store the value into the value structure */
	switch(attr->type)
	{
		case PAPI_STRING:
			sc = papiAttributeListAddString(attrs, PAPI_ATTR_APPEND, attr->name, value_str);
  			break;

		case PAPI_INTEGER:
  			int_val = atoi(value_str);
  			if ((int_val == 0) && (value_str[0] != '0'))
				sc = PAPI_BAD_REQUEST;
			else
				sc = papiAttributeListAddInteger(attrs, PAPI_ATTR_APPEND, attr->name, int_val);
  			break;

		case PAPI_DATETIME:
			time(&ltime);
			p_tm = localtime(&ltime);
			memcpy(&tm, p_tm, sizeof(struct tm));
			switch(strlen(value_str))
			{
			case 4:
				tm.tm_hour = papi_get_str_digits_value(value_str, 2);
				tm.tm_min = papi_get_str_digits_value(&(value_str[2]), 2);
				break;
			case 6:
				tm.tm_hour = papi_get_str_digits_value(value_str, 2);
				tm.tm_min = papi_get_str_digits_value(&(value_str[2]), 2);
				tm.tm_sec = papi_get_str_digits_value(&(value_str[4]), 2);
				break;
			case 8:
				tm.tm_year = papi_get_str_digits_value(value_str, 4) - 1900;
				tm.tm_mon = papi_get_str_digits_value(&(value_str[4]), 2) - 1;
				tm.tm_mday = papi_get_str_digits_value(&(value_str[6]), 2);
				break;
			case 12:
				tm.tm_year = papi_get_str_digits_value(value_str, 4) - 1900;
				tm.tm_mon = papi_get_str_digits_value(&(value_str[4]), 2) - 1;
				tm.tm_mday = papi_get_str_digits_value(&(value_str[6]), 2);
				tm.tm_hour = papi_get_str_digits_value(&(value_str[8]), 2);
				tm.tm_min = papi_get_str_digits_value(&(value_str[10]), 2);
				break;
			case 14:
				tm.tm_year = papi_get_str_digits_value(value_str, 4) - 1900;
				tm.tm_mon = papi_get_str_digits_value(&(value_str[4]), 2) - 1;
				tm.tm_mday = papi_get_str_digits_value(&(value_str[6]), 2);
				tm.tm_hour = papi_get_str_digits_value(&(value_str[8]), 2);
				tm.tm_min = papi_get_str_digits_value(&(value_str[10]), 2);
				tm.tm_sec = papi_get_str_digits_value(&(value_str[12]), 2);
				break;
			default:
				sc = PAPI_BAD_REQUEST;
				break;
			}
			if (sc == PAPI_OK)
			{
				ltime = mktime(&tm);
				sc = papiAttributeListAddDatetime(attrs, PAPI_ATTR_APPEND, attr->name, ltime);
			}
			break;

		case PAPI_RANGE:
			low = value_str;
			high = NULL;
			for(i=0; (value_str[i] != '\0') && (value_str[i] != '-'); i++) ;
			if (value_str[i] == '\0')
			{
				sc = PAPI_BAD_REQUEST;
				break;
			}
			value_str[i] = '\0';
			high = &value_str[i+1];

			sc = papiAttributeListAddRange(attrs, PAPI_ATTR_APPEND, attr->name, low, high);
			break;

		case PAPI_RESOLUTION:
			len = strlen(value_str);
			if (len < 4)
			{
				sc = PAPI_BAD_REQUEST;
				break;
			}
			if ((value_str[len-2] != 'p') || (value_str[len-3] != 'd'))
			{
				sc = PAPI_BAD_REQUEST;
				break;
			}
			if (value_str[len-1] == 'i')
				units = PAPI_RES_PER_INCH;
			else if (value_str[len-1] == 'c')
				units = PAPI_RES_PER_CM;
			else
			{
				sc = PAPI_BAD_REQUEST;
				break;
			}

			value_str[len-3] = '\0';
			for(i=0; (value_str[i] != '\0')&&(value_str[i] != 'x'); i++) ;
			if (value_str[i] == '\0')
			{
				xres = atoi(value_str);
				yres = xres;
			}
			else
			{
				yres = atoi(&(value_str[i+1]));
				value_str[i] = '\0';
				xres = atoi(value_str);
			}

			sc = papiAttributeListAddResolution(attrs, PAPI_ATTR_APPEND, attr->name,
				               xres, yres, units);
			break;
			
		default:
			sc = PAPI_BAD_REQUEST;
			break;
	}

	if (sc != PAPI_OK)
	{
		papi_dbg_printf("Syntax error in attributes at line %d. sc=%d\n",
                        __LINE__, sc);
		return(0);
	}

	return(1);
}

/**********************************************************************
 *  papi_parse_values
 *********************************************************************/
int papi_parse_values(char** attr_str, papi_attribute_t* attr,
                     int add_flags, papi_attribute_t*** attrs)
{
	int rc = 1;
	int i;
	char value_str[IPP_MAX_NAME];

	/* Skip over preceding whitespace */
	papi_skip_whitespace(attr_str);
	if ((**attr_str == '\0') || (**attr_str == ';') || (**attr_str == '='))
	{
		papi_dbg_printf("Syntax error in attributes at line %d\n", __LINE__);
		papi_dbg_printf("*attr_str: \"%s\"\n", *attr_str);
		return(0);
	}

	/* Process all the specified values */
	i = 0;
	while((**attr_str != '\0') && (**attr_str != ' ') &&
          (**attr_str != '\n') && (rc))
    {
    	if (**attr_str == ',')
    		(*attr_str)++;
		rc = papi_parse_value(i++, attr_str, add_flags, attr, attrs);
	}

	return(rc);
}

/**********************************************************************
 *  papi_get_value_type
 *********************************************************************/
typedef struct
{
	char* 		attr;
	papi_attribute_value_type_t	type;
} attr_value_type_map_t;

static attr_value_type_map_t ATTR_VALUE_TYPE_MAP[] =
{
	{ "attributes-charset",				PAPI_STRING },
	{ "attributes-natural-language",	PAPI_STRING },
	{ "charset-configured",				PAPI_STRING },
	{ "charset-supported",				PAPI_STRING },
	{ "color-supported",				PAPI_BOOLEAN },
	{ "compression-supported",			PAPI_STRING },
	{ "copies",							PAPI_INTEGER },
	{ "document-format",				PAPI_STRING },
	{ "document-format-default",		PAPI_STRING },
	{ "document-format-supported",		PAPI_STRING },
	{ "finishings",						PAPI_INTEGER },
	{ "generated-natural-language-supported",PAPI_STRING },
	{ "job-hold-until",					PAPI_STRING },
	{ "job-id",							PAPI_INTEGER },
	{ "job-impressions",				PAPI_INTEGER },
	{ "job-impressions-completed",		PAPI_INTEGER },
	{ "job-impressions-supported",		PAPI_RANGE },
	{ "job-k-octets",					PAPI_INTEGER },
	{ "job-k-octets-processed",			PAPI_INTEGER },
	{ "job-k-octets-supported",			PAPI_RANGE },
	{ "job-media-sheets",				PAPI_INTEGER },
	{ "job-media-sheets-completed",		PAPI_INTEGER },
	{ "job-media-sheets-supported",		PAPI_RANGE },
	{ "job-message-from-operator",		PAPI_STRING },
	{ "job-more-info",					PAPI_STRING },
	{ "job-name",						PAPI_STRING },
	{ "job-originating-user-name",		PAPI_STRING },
	{ "job-printer-uri",				PAPI_STRING },
	{ "job-priority",					PAPI_INTEGER },
	{ "job-sheets",						PAPI_STRING },
	{ "job-state",						PAPI_INTEGER },
	{ "job-state-message",				PAPI_STRING },
	{ "job-state-reasons",				PAPI_STRING },
	{ "job-uri",						PAPI_STRING },
	{ "media", 							PAPI_STRING },
	{ "message", 						PAPI_STRING },
	{ "multiple-operation-time-out",	PAPI_INTEGER },
	{ "multiple-document-handling",		PAPI_STRING },
	{ "natural-language-configured",	PAPI_STRING },
	{ "number-of-documents", 			PAPI_INTEGER },
	{ "number-of-intervening-jobs",		PAPI_INTEGER },
	{ "number-up",						PAPI_INTEGER },
	{ "operations-supported",			PAPI_INTEGER },
	{ "orientation-requested",			PAPI_INTEGER },
	{ "output-device-assigned",			PAPI_STRING },
	{ "page-ranges",					PAPI_RANGE },
	{ "pdl-override-supported",			PAPI_STRING },
	{ "print-quality",					PAPI_INTEGER },
	{ "printer-current-time",			PAPI_DATETIME },
	{ "printer-driver-installer",		PAPI_STRING },
	{ "printer-info",					PAPI_STRING },
	{ "printer-is-accepting-jobs",		PAPI_BOOLEAN },
	{ "printer-location",				PAPI_STRING },
	{ "printer-make-and-model",			PAPI_STRING },
	{ "printer-message-from-operator",	PAPI_STRING },
	{ "printer-more-info",				PAPI_STRING },
	{ "printer-more-info-manufacturer",	PAPI_STRING },
	{ "printer-name",					PAPI_STRING },
	{ "printer-resolution",				PAPI_RESOLUTION },
	{ "printer-state",					PAPI_INTEGER },
	{ "printer-state-message",			PAPI_STRING },
	{ "printer-state-reasons",			PAPI_STRING },
	{ "printer-up-time",				PAPI_INTEGER },
	{ "printer-uri",					PAPI_STRING },
	{ "printer-uri-supported",			PAPI_STRING },
	{ "queued-job-count",				PAPI_INTEGER },
	{ "reference-uri-schemes-supported",PAPI_STRING },
	{ "requesting-user-name",			PAPI_STRING },
	{ "sides",							PAPI_STRING },
	{ "time-at-completed",				PAPI_INTEGER },
	{ "time-at-creation",				PAPI_INTEGER },
	{ "time-at-processing",				PAPI_INTEGER },
	{ "uri-security-supported",			PAPI_STRING },
	/* End of table */
	{ NULL, PAPI_STRING } /* NOTE: we guess "string" if not found */
};

papi_attribute_value_type_t papi_get_value_type(char* attr_name)
{
	int i;
		
	for(i=0; ATTR_VALUE_TYPE_MAP[i].attr != NULL; i++)	
	{
		if ( !strcmp(ATTR_VALUE_TYPE_MAP[i].attr, attr_name) )
			break;
	}

	return( ATTR_VALUE_TYPE_MAP[i].type );
}

/**********************************************************************
 *  papi_parse_attr_add
 *********************************************************************/
int papi_parse_attr_add(char** attr_str, char* attr_name, papi_attribute_t*** attrs)
{
	int i;
	papi_status_t sc;
	papi_attribute_value_type_t attr_type;

	/* Parse off the attribute name */
	for(i=0;
	    ( (**attr_str != ' ') &&
		  (**attr_str != '=') &&
		  (**attr_str != '\0') &&
		  (i < IPP_MAX_NAME-1) );
		i++)
	{	
		attr_name[i] = tolower(**attr_str);			
		(*attr_str)++;
	}
	if ((i == IPP_MAX_NAME-1)||(**attr_str == '\0'))
		return(0);
	attr_name[i] = '\0';

	/* Determine the type of attribute this is */
	attr_type = papi_get_value_type(attr_name);

	/* Add it with no values */
	sc = papiAttributeListAddValue(attrs, PAPI_ATTR_EXCL, attr_name, attr_type, NULL);
	if (sc != PAPI_OK)
	{
		papi_dbg_printf("papiAttributeListAdd failed (rc=%d): %s\n", sc, papiStatusString(sc));
		return(0);
	}

	return(1);
}

/**********************************************************************
 *  papi_parse_attr
 *********************************************************************/
int papi_parse_attr(char** attr_str, int add_flags, papi_attribute_t*** attrs)
{
	papi_attribute_t* attr = NULL;	
	int i;
	int num_values = 0;
	char attr_name[IPP_MAX_NAME];

	/* Skip over preceding whitespace */
	papi_skip_whitespace(attr_str);
	if (**attr_str == '\0')
		return(1);
	if (**attr_str == ';')
	{
		(*attr_str)++;
		return(1);
	}

	/* First pass: determine the number of values */
	num_values = papi_parse_attr_get_value_count(*attr_str);

	/* Add the new attribute */
	if ( !papi_parse_attr_add(attr_str, attr_name, attrs) )
		return(0);
	
	/* Get to the values */
	papi_skip_whitespace(attr_str);
	if (**attr_str != '=')
	{
		papi_dbg_printf("Input attribute has no value\n");
		return(0);
	}
	(*attr_str)++;

	/* Process all the specified values */
	attr = papiAttributeListFind(*attrs, attr_name);
	if (attr == NULL)
	{
		papi_dbg_printf("Internal error at line %d\n", __LINE__);
		return(0);
	}
	return( papi_parse_values(attr_str, attr, add_flags, attrs) );
}

papi_status_t papi_attribute_to_string(papi_attribute_t* attr, char* delim,
           char* buffer, size_t buflen)
{
	papi_status_t rc = PAPI_OK;
	int i;
	int k;
	int yr;
	size_t currlen;
	char* val_str;
	char  tmp_buff[100];
	struct tm* p_tm;
	struct tm tm;

	if (attr == NULL)
		return(PAPI_BAD_REQUEST);

	/* Add the delimiter (if not the first one) */
	if (buffer[0] != '\0')
	{
		currlen = strlen(buffer) + 1;
		currlen += strlen(delim);
		if (currlen > buflen) return(PAPI_BAD_REQUEST);
		strcat(buffer, delim);
	}
	else
	{
		currlen = 0;
	}

	/* Add the attribute name and equals sign */
	currlen += strlen(attr->name);
	currlen++;
	if (currlen > buflen) return(PAPI_BAD_REQUEST);
	strcat(buffer, attr->name);	
	strcat(buffer, "=");	

	/* Add the value string(s) */
	for(i=0; attr->values[i] != NULL; i++)
	{
		if (i>0)
		{
			currlen++;
			if (currlen > buflen) return(PAPI_BAD_REQUEST);
			strcat(buffer, ",");
		}

		switch(attr->type)
		{
		case PAPI_STRING:
			currlen++;
			if (currlen > buflen) return(PAPI_BAD_REQUEST);
			strcat(buffer, "\"");
            for(k=0; attr->values[i]->string[k] != '\0'; k++)
            {
            	if ( attr->values[i]->string[k] != '\"' )
            	{	
					currlen++;
					if (currlen > buflen) return(PAPI_BAD_REQUEST);
					buffer [currlen-2] = attr->values[i]->string[k];
					buffer [currlen-1] = '\0';
				}
				else
            	{	
					currlen += 2;
					if (currlen > buflen) return(PAPI_BAD_REQUEST);
					buffer [currlen-3] = '\\';
					buffer [currlen-2] = '"';
					buffer [currlen-1] = '\0';
				}
				
            }
			currlen++;
			if (currlen > buflen) return(PAPI_BAD_REQUEST);
			strcat(buffer, "\"");
			break;
		case PAPI_BOOLEAN:
			if (attr->values[i]->boolean)
				val_str = "true";
			else
				val_str = "false";
			currlen += strlen(val_str);
			if (currlen > buflen) return(PAPI_BAD_REQUEST);
			strcat(buffer, val_str);
        	break;
		case PAPI_INTEGER:
			sprintf(tmp_buff, "%d", attr->values[i]->integer);
			currlen += strlen(tmp_buff);
			if (currlen > buflen) return(PAPI_BAD_REQUEST);
			strcat(buffer, tmp_buff);
        	break;
		case PAPI_RANGE:
			sprintf(tmp_buff, "%d-%d", attr->values[i]->range.lower,
                       attr->values[i]->range.upper);
			currlen += strlen(tmp_buff);
			if (currlen > buflen) return(PAPI_BAD_REQUEST);
			strcat(buffer, tmp_buff);
        	break;
		case PAPI_DATETIME:
			p_tm = localtime(&(attr->values[i]->datetime));
			memcpy(&tm, p_tm, sizeof(struct tm));
			sprintf(tmp_buff, "%d%02d%02d%02d%02d%02d",
					(tm.tm_year+1900),
					(tm.tm_mon+1),
					tm.tm_mday,
					tm.tm_hour,
					tm.tm_min,
					tm.tm_sec);
			currlen += strlen(tmp_buff);
			if (currlen > buflen) return(PAPI_BAD_REQUEST);
			strcat(buffer, tmp_buff);
			break;
		case PAPI_RESOLUTION:
			if (attr->values[i]->resolution.units == PAPI_RES_PER_INCH)
				sprintf(tmp_buff, "%dx%ddpi", attr->values[i]->resolution.xres,
                       attr->values[i]->resolution.yres);
            else
				sprintf(tmp_buff, "%dx%ddpc", attr->values[i]->resolution.xres,
                       attr->values[i]->resolution.yres);
			currlen += strlen(tmp_buff);
			if (currlen > buflen) return(PAPI_BAD_REQUEST);
			strcat(buffer, tmp_buff);
			break;
		case PAPI_COLLECTION:
			/* ?????????????????? */
			break;
		default:
			printf("unrecognized attr type=0x%x", attr->type);
			break;
		}
	}

	return(PAPI_OK);
}


/**********************************************************************
 * papi_dup_cups_attr
 *********************************************************************/
ipp_attribute_t* papi_dup_cups_attr(ipp_attribute_t* attr, ipp_t* dup_req)
{
	int i;
	ipp_attribute_t* new_attr = NULL;

 	new_attr = papi_add_cups_attr(dup_req, attr->num_values);
	new_attr->name = strdup(attr->name);
	new_attr->group_tag = attr->group_tag;
	new_attr->value_tag = attr->value_tag;

 	for(i=0; i < attr->num_values; i++)
 	{
 		switch(attr->value_tag)
 		{
		case IPP_TAG_CHARSET:
		case IPP_TAG_KEYWORD:
		case IPP_TAG_LANGUAGE:
		case IPP_TAG_MEMBERNAME:
		case IPP_TAG_MIMETYPE:
		case IPP_TAG_NAME:
		case IPP_TAG_NAMELANG:
		case IPP_TAG_NOVALUE:
		case IPP_TAG_STRING:
		case IPP_TAG_TEXT:
		case IPP_TAG_TEXTLANG:
		case IPP_TAG_URI:
		case IPP_TAG_URISCHEME:
			if (attr->values[i].string.charset)
				new_attr->values[i].string.charset = strdup(attr->values[i].string.charset);
			else
				new_attr->values[i].string.charset = NULL;
			if (attr->values[i].string.text)
				new_attr->values[i].string.text = strdup(attr->values[i].string.text);
			else
				new_attr->values[i].string.text = NULL;
  			break;

		case IPP_TAG_ENUM:
		case IPP_TAG_INTEGER:
			new_attr->values[i].integer = attr->values[i].integer;
			break;

		case IPP_TAG_BOOLEAN:
			new_attr->values[i].boolean = attr->values[i].boolean;
			break;

		case IPP_TAG_DATE:
			memcpy(new_attr->values[i].date, attr->values[i].date, 11);
			break;

		case IPP_TAG_RANGE:
			new_attr->values[i].range.lower = attr->values[i].range.lower;
			new_attr->values[i].range.upper = attr->values[i].range.upper;
			break;

		case IPP_TAG_RESOLUTION:
			new_attr->values[i].resolution.xres = attr->values[i].resolution.xres;
			new_attr->values[i].resolution.yres = attr->values[i].resolution.yres;
			new_attr->values[i].resolution.units = attr->values[i].resolution.units;
			break;
			
		default:
			papi_dbg_printf("unrecognized tag: 0x%x\n", attr->value_tag);
		}
 	}

 	return(new_attr);
}

/**********************************************************************
 * papi_dup_cups_request
 *********************************************************************/
ipp_t* papi_dup_cups_request(ipp_t* req)
{
	ipp_t* dup_req = NULL;
	ipp_attribute_t* attr = NULL;
	ipp_attribute_t* new_attr = NULL;

	dup_req = ippNew();

	memcpy(dup_req, req, sizeof(ipp_t));

	dup_req->attrs = NULL;
	dup_req->last = NULL;
	dup_req->current = NULL;

	for(attr=req->attrs; attr != NULL; attr=attr->next)
		papi_dup_cups_attr(attr, dup_req);

	return(dup_req);
}

/*---------------------------------------------------------------------
 * Attributes API
 *-------------------------------------------------------------------*/

/**********************************************************************
 * papiAttributeListAdd
 *********************************************************************/
papi_status_t papiAttributeListAddValue(
          papi_attribute_t*** attrs,
          int add_flags,
          char* name,
          papi_attribute_value_type_t type,
          papi_attribute_value_t* value )
{
    int i;
    int attr_count = 0;
    papi_attribute_t* attr;
    int add_attr = 0;

	papi_dbg_printf(">papiAttributeListAddValue(%s)\n", name);

	if (name == NULL)
		PAPI_RETURN("papiAttributeListAdd", NULL, PAPI_BAD_REQUEST);

	/*-----------------------------------------------------------------
	 * See if this attribute is already present and handle accordingly
	 *---------------------------------------------------------------*/
	if (*attrs != NULL)
	{
		for(i=0; (*attrs)[i] != NULL; i++)
		{
			if ( !strcmp((*attrs)[i]->name, name) )
			    break;
		}
		if ((*attrs)[i])
		{
			papi_dbg_printf("already exists in list\n");
			if (add_flags & PAPI_ATTR_EXCL)
				PAPI_RETURN("papiAttributeListAdd", NULL, PAPI_BAD_REQUEST);
			if (add_flags & PAPI_ATTR_REPLACE)
			{
				papi_dbg_printf("replacing values...\n");
				papi_attr_free((*attrs)[i]);
				(*attrs)[i] = malloc(sizeof(papi_attribute_t));
				if ((*attrs)[i] == NULL)
					PAPI_RETURN("papiAttributeListAdd", NULL, PAPI_INTERNAL_ERROR);
				attr = (*attrs)[i];
				attr->name = strdup(name);
				attr->type = type;
				attr->values = NULL;
			}
			else if (add_flags & PAPI_ATTR_APPEND)
			{
				papi_dbg_printf("adding value...\n");
				attr = (*attrs)[i];
			}
		}
		else
		{
			papi_dbg_printf("adding to existing list\n");
			add_attr = 1;
		}
	}
	else
	{
		papi_dbg_printf("adding to new list\n");
		add_attr = 1;
	}

	if (add_attr)
	{
		/*-----------------------------------------------------------------
		 * Attribute to be added: ensure there is space in the attribute list
		 *---------------------------------------------------------------*/
		if (*attrs == NULL)
		{
			/* First attribute in the list... */
			*attrs = malloc(sizeof(papi_attribute_t*) * 2);
			if (*attrs == NULL)
				PAPI_RETURN("papiAttributeListAdd", NULL, PAPI_INTERNAL_ERROR);
			i = 0;
		}
		else
		{
			/* Extending existing list... */
			for(i=0; (*attrs)[i] != NULL; i++)
			   attr_count++;
			*attrs = realloc(*attrs, sizeof(papi_attribute_t*) * (attr_count+2));
			if (*attrs == NULL)
				PAPI_RETURN("papiAttributeListAdd", NULL, PAPI_INTERNAL_ERROR);
		}

		/* Set up the new attribute */
		(*attrs)[i+1] = NULL;
		(*attrs)[i] = malloc(sizeof(papi_attribute_t));
		if ((*attrs)[i] == NULL)
			PAPI_RETURN("papiAttributeListAdd", NULL, PAPI_INTERNAL_ERROR);
		attr = (*attrs)[i];
		attr->name = strdup(name);
		attr->type = type;
		attr->values = NULL;
	}

	/* Handle adding an attribute without a value */
	if (value == NULL)
		PAPI_RETURN("papiAttributeListAdd (no values added)", NULL, PAPI_OK);

	/* Set i to the index for the new value */
	if ((attr->values == NULL) || (attr->values[0] == NULL))
	{
		i = 0;
	}
	else
	{
		for(i=1; attr->values[i] != NULL; i++) ;
	}

	/* Set up memory for the new value */
	if (attr->values == NULL)
		attr->values = malloc(sizeof(papi_attribute_value_t*) * (i+2));
	else
		attr->values = realloc(attr->values, sizeof(papi_attribute_value_t*) * (i+2));
	attr->values[i] = malloc( sizeof(papi_attribute_value_t) );
	attr->values[i+1] = NULL;		

	/* Add the new value */
	switch(attr->type)
	{
	case PAPI_STRING:
		attr->values[i]->string = strdup(value->string);
		break;
	case PAPI_INTEGER:
		attr->values[i]->integer = value->integer;
		break;
	case PAPI_BOOLEAN:
		attr->values[i]->boolean = value->boolean;
		break;
	case PAPI_RANGE:
		attr->values[i]->range.lower = value->range.lower;
		attr->values[i]->range.upper = value->range.upper;
		break;
	case PAPI_RESOLUTION:
		attr->values[i]->resolution.xres = value->resolution.xres;
		attr->values[i]->resolution.yres = value->resolution.yres;
		attr->values[i]->resolution.units = value->resolution.units;
		break;
	case PAPI_DATETIME:
		memcpy(&(attr->values[i]->datetime), &(value->datetime), sizeof(time_t));
		break;
	case PAPI_COLLECTION:
		attr->values[i]->collection = value->collection;
		break;
	default:
		PAPI_RETURN("papiAttributeListAdd", NULL, PAPI_BAD_REQUEST);
	}

	PAPI_RETURN("papiAttributeListAdd", NULL, PAPI_OK);
}

/**********************************************************************
 * papiAttributeListAddString
 *********************************************************************/
papi_status_t papiAttributeListAddString(
          papi_attribute_t*** attrs,
          int add_flags,
          char* name,
          char* value)
{
	papi_status_t rc;
	papi_attribute_value_t attr_value;
	if (value)
	{
    	attr_value.string = value;
		rc = papiAttributeListAddValue(attrs, add_flags, name, PAPI_STRING, &attr_value);
	}
	else
	{
		rc = papiAttributeListAddValue(attrs, add_flags, name, PAPI_STRING, NULL);
	}
	return(rc);
}

/**********************************************************************
 * papiAttributeListAddInteger
 *********************************************************************/
papi_status_t papiAttributeListAddInteger(
          papi_attribute_t*** attrs,
          int add_flags,
          char* name,
          int value)
{
	papi_attribute_value_t attr_value;
   	attr_value.integer = value;
	return( papiAttributeListAddValue(attrs, add_flags, name, PAPI_INTEGER, &attr_value) );
}

/**********************************************************************
 * papiAttributeListAddBoolean
 *********************************************************************/
papi_status_t papiAttributeListAddBoolean(
          papi_attribute_t*** attrs,
          int add_flags,
          char* name,
          char value)
{
	papi_attribute_value_t attr_value;
   	attr_value.boolean = value;
	return( papiAttributeListAddValue(attrs, add_flags, name, PAPI_BOOLEAN, &attr_value) );
}

/**********************************************************************
 * papiAttributeListAddRange
 *********************************************************************/
papi_status_t papiAttributeListAddRange(
          papi_attribute_t*** attrs,
          int add_flags,
          char* name,
          int lower,
          int upper )
{
	papi_attribute_value_t attr_value;
   	attr_value.range.lower = lower;
   	attr_value.range.upper = upper;
	return( papiAttributeListAddValue(attrs, add_flags, name, PAPI_RANGE, &attr_value) );
}

/**********************************************************************
 * papiAttributeListAddResolution
 *********************************************************************/
papi_status_t papiAttributeListAddResolution(
          papi_attribute_t*** attrs,
          int add_flags,
          char* name,
          int xres,
          int yres,
          papi_resolution_unit_t units )
{
	papi_attribute_value_t attr_value;
   	attr_value.resolution.xres = xres;
   	attr_value.resolution.yres = yres;
   	attr_value.resolution.units = units;
	return( papiAttributeListAddValue(attrs, add_flags, name, PAPI_RESOLUTION, &attr_value) );
}

/**********************************************************************
 * papiAttributeListAddDatetime
 *********************************************************************/
papi_status_t papiAttributeListAddDatetime(
          papi_attribute_t*** attrs,
          int add_flags,
          char* name,
          time_t date_time )
{
	papi_attribute_value_t attr_value;
   	memcpy(&attr_value.datetime, &date_time, sizeof(time_t));
	return( papiAttributeListAddValue(attrs, add_flags, name, PAPI_DATETIME, &attr_value) );
}

/**********************************************************************
 * papiAttributeListAddCollection
 *********************************************************************/
papi_status_t papiAttributeListAddCollection(
          papi_attribute_t*** attrs,
          int add_flags,
          char* name,
          papi_attribute_t** collection )
{
	papi_attribute_value_t attr_value;
   	attr_value.collection = collection;
	return( papiAttributeListAddValue(attrs, add_flags, name, PAPI_COLLECTION, &attr_value) );
}

/**********************************************************************
 * papiAttributeDelete
 *********************************************************************/
papi_status_t papiAttributeDelete(
          papi_attribute_t*** attrs,
          char* name )
{
	int i;
	papi_status_t rc = PAPI_NOT_FOUND;
	papi_attribute_t* attr;

	if (attrs != NULL)
	{
		/* Find the attribute */
		for(i=0; (*attrs)[i] != NULL; i++)
		{
			if ( !strcmp((*attrs)[i]->name, name) )
				break;
		}

		if (attrs[i] != NULL)
		{
			rc = PAPI_OK;
			/* Free the attribute memory */
			papi_attr_free((*attrs)[i]);
			/* Move everything in the list down one */
			for( ; (*attrs)[i] != NULL; i++)
				(*attrs)[i] = (*attrs)[i+1];
		}
	}

	PAPI_RETURN("papiAttributeDelete", NULL, rc);
}

/**********************************************************************
 * papiAttributeListGetValue
 *********************************************************************/
papi_status_t papiAttributeListGetValue(
          papi_attribute_t** attrs,
                void** iterator,
          char* name,
          papi_attribute_value_type_t type,
                papi_attribute_value_t** value )
{
	int  i;
	int  val_i = 0;
	int* val_index = &val_i;

	papi_dbg_printf(">papiAttributeListGetValue(%s)\n", name);

	if (iterator != NULL)
		val_index = (int*)iterator;

	if ((attrs == NULL) || (name == NULL) || (value == NULL))
		PAPI_RETURN("papiAttributeListGetValue", NULL, PAPI_BAD_REQUEST);

	

	/* Search for the attribute */
	for(i=0; attrs[i] != NULL; i++)
	{
		if ( !strcmp(attrs[i]->name, name) )
			break;
	}


	if (attrs[i] == NULL)
		PAPI_RETURN("papiAttributeListGetValue", NULL, PAPI_NOT_FOUND);

	if (type != attrs[i]->type)
		PAPI_RETURN("papiAttributeListGetValue", NULL, PAPI_BAD_REQUEST);

	/* Check for requested value */
	if ((attrs[i]->values == NULL) || (attrs[i]->values[*val_index] == NULL))
		PAPI_RETURN("papiAttributeListGetValue", NULL, PAPI_NOT_FOUND);

	/* Get the pointer to  the value */
	*value = attrs[i]->values[*val_index];
 	(*val_index)++;

	PAPI_RETURN("papiAttributeListGetValue", NULL, PAPI_OK);
}

/**********************************************************************
 * papiAttributeListGetString
 *********************************************************************/
papi_status_t papiAttributeListGetString(
          papi_attribute_t** attrs,
                void** iterator,
          char* name,
                char** value )
{
	papi_status_t rc;
	papi_attribute_value_t* attr_value;

	if (value == NULL)
		return(PAPI_BAD_REQUEST);

	*value = NULL;
	rc = papiAttributeListGetValue(attrs, iterator, name, PAPI_STRING, &attr_value);
	if (rc == PAPI_OK)
		*value = attr_value->string;

	return(rc);
}

/**********************************************************************
 * papiAttributeListGetInteger
 *********************************************************************/
papi_status_t papiAttributeListGetInteger(
          papi_attribute_t** attrs,
                void** iterator,
          char* name,
                int* value )
{
	papi_status_t rc;
	papi_attribute_value_t* attr_value;

	if (value == NULL)
		return(PAPI_BAD_REQUEST);

	rc = papiAttributeListGetValue(attrs, iterator, name, PAPI_INTEGER, &attr_value);
	if (rc == PAPI_OK)
		*value = attr_value->integer;

	return(rc);
}

/**********************************************************************
 * papiAttributeListGetBoolean
 *********************************************************************/
papi_status_t papiAttributeListGetBoolean(
          papi_attribute_t** attrs,
                void** iterator,
          char* name,
                char* value )
{
	papi_status_t rc;
	papi_attribute_value_t* attr_value;

	if (value == NULL)
		return(PAPI_BAD_REQUEST);

	rc = papiAttributeListGetValue(attrs, iterator, name, PAPI_BOOLEAN, &attr_value);
	if (rc == PAPI_OK)
		*value = attr_value->boolean;

	return(rc);
}

/**********************************************************************
 * papiAttributeListGetRange
 *********************************************************************/
papi_status_t papiAttributeListGetRange(
          papi_attribute_t** attrs,
                void** iterator,
          char* name,
                int* lower,
                int* upper )
{
	papi_status_t rc;
	papi_attribute_value_t* attr_value;

	if ((lower == NULL) || (upper == NULL))
		return(PAPI_BAD_REQUEST);

	rc = papiAttributeListGetValue(attrs, iterator, name, PAPI_RANGE, &attr_value);
	if (rc == PAPI_OK)
	{
		*lower = attr_value->range.lower;
		*upper = attr_value->range.upper;
	}

	return(rc);
}

/**********************************************************************
 * papiAttributeListGetResolution
 *********************************************************************/
papi_status_t papiAttributeListGetResolution(
          papi_attribute_t** attrs,
                void** iterator,
          char* name,
                int* xres,
                int* yres,
                papi_resolution_unit_t* units )
{
	papi_status_t rc;
	papi_attribute_value_t* attr_value;

	if ((xres == NULL) || (yres == NULL) || (units == NULL))
		return(PAPI_BAD_REQUEST);

	rc = papiAttributeListGetValue(attrs, iterator, name, PAPI_RESOLUTION, &attr_value);
	if (rc == PAPI_OK)
	{
		*xres = attr_value->resolution.xres;
		*yres = attr_value->resolution.yres;
		*units = attr_value->resolution.units;
	}

	return(rc);
}

/**********************************************************************
 * papiAttributeListGetDatetime
 *********************************************************************/
papi_status_t papiAttributeListGetDatetime(
          papi_attribute_t** attrs,
                void** iterator,
          char* name,
                time_t* date_time)
{
	papi_status_t rc;
	papi_attribute_value_t* attr_value;

	if (date_time == NULL)
		return(PAPI_BAD_REQUEST);

	rc = papiAttributeListGetValue(attrs, iterator, name, PAPI_DATETIME, &attr_value);
	if (rc == PAPI_OK)
		memcpy(date_time, &(attr_value->datetime), sizeof(time_t));

	return(rc);
}

/**********************************************************************
 * papiAttributeListGetCollection
 *********************************************************************/
papi_status_t papiAttributeListGetCollection(
          papi_attribute_t** attrs,
                void** iterator,
          char* name,
                papi_attribute_t*** collection)
{
	papi_status_t rc;
	papi_attribute_value_t* attr_value;

	if (collection == NULL)
		return(PAPI_BAD_REQUEST);

	rc = papiAttributeListGetValue(attrs, iterator, name, PAPI_COLLECTION, &attr_value);
	if (rc == PAPI_OK)
		*collection = attr_value->collection;

	return(rc);
}

/**********************************************************************
 * papiAttributeListFree
 *********************************************************************/
void papiAttributeListFree(papi_attribute_t** attrs )
{
	int i;	
	int k;	

	if (attrs)
	{
		/* Process all of the attributes */
		for(i=0; attrs[i] != NULL; i++)
		{
			/* Process all of this attribute's values */
			for(k=0; attrs[i]->values[k] != NULL; k++)
			{
				if (attrs[i]->type == PAPI_STRING)
					free(attrs[i]->values[k]->string);
				free(attrs[i]->values[k]);
			}
			if (attrs[i]->name)
				free(attrs[i]->name);
			free(attrs[i]);
		}
		free(attrs);
	}
}

/**********************************************************************
 * papiAttributeListFind
 *********************************************************************/
papi_attribute_t* papiAttributeListFind(
          papi_attribute_t** attrs,
          char*              name )
{
	papi_attribute_t* attr = NULL;
	int i;

	if (attrs)
	{
		/* Process all of the attributes */
		for(i=0; attrs[i] != NULL; i++)
		{
			if ( !strcmp(attrs[i]->name, name) )
			{
				attr = attrs[i];
				break;
			}
		}
	}

	return(attr);
}

/**********************************************************************
 * papiAttributeListGetNext
 *********************************************************************/
papi_attribute_t* papiAttributeListGetNext(
          papi_attribute_t** attrs,
                void**             iterator )
{
	int i;
	if ((iterator == NULL) || (attrs == NULL))
		return(NULL);

	for(i=0; attrs[i] != NULL; i++)
		if (i == (int)*iterator)
			break;

	/* ((int)*iterator)++; */
	{
		int *i = *iterator;
		(*i)++;
	}

	return( attrs[i] );
}

/**********************************************************************
 * papiAttributeListFromString
 *********************************************************************/
papi_status_t papiAttributeListFromString(
          papi_attribute_t*** attrs,
          int                add_flags,
          char*              buffer )
{
	char* p_attr = (char*)buffer;

	if (p_attr != NULL)
	{
		/* Process all the attrs in the string */
		while(*p_attr != '\0')
		{
			/* Process this attribute */
			if ( !papi_parse_attr(&p_attr, add_flags, attrs) )			
	              return(PAPI_BAD_REQUEST);
		}
	}
	
	return(PAPI_OK);
}

/**********************************************************************
 * papiAttributeListToString
 *********************************************************************/
papi_status_t papiAttributeListToString(
          papi_attribute_t** attrs,
          char*              attr_delim,
                char*              buffer,
          size_t             buflen )
{
	papi_status_t rc = PAPI_OK;
	int i;

	if ((buffer == NULL) || (buflen < 1))
		return(PAPI_BAD_REQUEST);
	
	buffer[0] = '\0';

	if (attrs == NULL)
		return(PAPI_BAD_REQUEST);


	for(i=0; (attrs[i] != NULL) && (rc == PAPI_OK); i++)
		rc = papi_attribute_to_string(attrs[i], attr_delim, buffer, buflen);

	return(rc);
}
