/**********************************************************************
 *
 * Open Standard Print API (PAPI) - CUPS-specific Functions
 *
 * Copyright (c) 2002 International Business Machines Corp.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "papi.h"
#include "papi_int.h"

/*---------------------------------------------------------------------
 *                     Local Functions
 *-------------------------------------------------------------------*/

/**********************************************************************
 * papi_cups_get_printers
 *********************************************************************/
int papi_cups_get_printers(papi_int_service_t* handle, char*** printers)
{
	int count = 0;
	ipp_attribute_t* cups_attr = NULL;
	ipp_value_t* cups_val = NULL;
	papi_dbg_printf(">papi_cups_get_printers\n");

	*printers = malloc(sizeof(char*));
	(*printers)[0] = NULL;

	if (handle == NULL)
	{
		papi_dbg_printf("<papi_cups_get_printers returning 0 at line %d\n", __LINE__);
		return(0);
	}
	if (handle->response == NULL)
	{
		papi_dbg_printf("<papi_cups_get_printers returning 0 at line %d\n", __LINE__);
		return(0);
	}

	/* Process the returned attributes */
	for(cups_attr=handle->response->attrs;
        cups_attr != NULL;
        cups_attr=cups_attr->next)
    {
    	if ( cups_attr->name && !strcmp(cups_attr->name, "printer-name") )
    	{
    		count++;
    		*printers = realloc(*printers, sizeof(char*) * (count+1));
    		(*printers)[count-1] = strdup(cups_attr->values[0].string.text);
    		(*printers)[count] = NULL;
    	}
	}

	papi_dbg_printf("<papi_cups_get_printers returning %d\n", count);
	return(count);
}

/*---------------------------------------------------------------------
 *
 *                     Public (API) Functions
 *
 *-------------------------------------------------------------------*/

/**********************************************************************
 * papiPrintersList
 *********************************************************************/
papi_status_t papiPrintersList(
                papi_service_t     handle,
          char*              requested_attrs[],
          papi_filter_t*     filter,
                papi_printer_t**   printers )
{
	papi_status_t rc;
	char resource[HTTP_MAX_URI];
	static char* all_attrs[] = { "all" }; /* Requested attributes */
	char** cups_printers;
	int i;
	int count;
	int tmp;
	papi_int_printer_t** int_printers;

	papi_init();
	papi_dbg_printf(">papiPrintersList\n");
	*printers = NULL;

	if ((handle == NULL) || (printers == NULL))
		PAPI_RETURN("papiPrintersList", handle, PAPI_BAD_REQUEST);

	/* Set up the request (NOTE: We can't use the cupsGetPrinters
     * convenience function because it doesn't support specifying
     * a filter) */
	rc = papi_set_up_request(INT_HANDLE(handle), "", PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papiPrintersList", handle, rc);

	if (requested_attrs && (requested_attrs[0] != NULL))
	{
		/* get requested attributes */
		papi_dbg_printf("add req for selected attributes\n");
		ippAddStrings(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_NAME,
	                "requested-attributes",
					papi_get_array_count(requested_attrs),
					NULL, requested_attrs);
	}

	/* Set the printer-type-mask if filtering */
	if (filter && (filter->type == PAPI_FILTER_BITMASK))
	{
		papi_dbg_printf("filtering with mask=0x%x, value=0x%x\n",
					filter->filter.bitmask.mask, filter->filter.bitmask.value);
		ippAddInteger(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_ENUM,
					"printer-type", filter->filter.bitmask.value);
		ippAddInteger(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_ENUM,
					"printer-type-mask", filter->filter.bitmask.mask);
	}

	INT_HANDLE(handle)->request->request.any.request_id = 1;
	INT_HANDLE(handle)->request->request.op.operation_id = CUPS_GET_PRINTERS;

	/* Get the list of printers from CUPS */
 	papi_do_request(handle, "/printers/", NULL);

	if (INT_HANDLE(handle)->response == NULL)
		PAPI_RETURN("papiPrintersList", handle, PAPI_BAD_REQUEST);
	rc = papi_get_result(INT_HANDLE(handle));
	if (rc != PAPI_OK)
		PAPI_RETURN("papiPrintersList", handle, rc);

	/* Get the number of and list of printers */
	count = papi_cups_get_printers(INT_HANDLE(handle), &cups_printers);

	/* Create the output printer list */
	int_printers = malloc(sizeof(papi_int_printer_t*) * (count+1));
	if (int_printers == NULL)
	{
		free(cups_printers);
		PAPI_RETURN("papiPrintersList", handle, PAPI_INTERNAL_ERROR);
	}
	int_printers[count] = NULL;
	*printers = (papi_printer_t)int_printers;

	/* For each printer on the list, query it */
	papi_clear_prev_request(INT_HANDLE(handle));
	for(i=0; i<count; i++)
	{
		rc = papiPrinterQuery(handle, cups_printers[i], requested_attrs, NULL, &(int_printers[i]));
		if (rc != PAPI_OK)
		{
			free(cups_printers);
			int_printers[i] = NULL;
			PAPI_RETURN("papiPrintersList", handle, rc);
		}
		
		papi_clear_prev_request(INT_HANDLE(handle));
	}

	free(cups_printers);
	PAPI_RETURN("papiPrintersList", handle, rc);
}
