/**********************************************************************
 *
 * Open Standard Print API (PAPI) - Internal Functions
 *
 * Copyright (c) 2002 International Business Machines Corp.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <locale.h>
#include "papi.h"
#include "papi_int.h"

/*---------------------------------------------------------------------
 * Static variables
 *-------------------------------------------------------------------*/

static int init_done = 0;

#define DEBUG_MASK_NONE   0
#define DEBUG_MASK_STDOUT 1
#define DEBUG_MASK_FILE   2
static int do_debug = 0;
static FILE* dbgf = NULL;

/* In the future, we may want to improve debugging in a
 * multithreaded environment... */
#if 0
#define DEBUG_MAX_THREADS 200
typedef struct
{
	int tid;
	int nest_level;
} dbg_thread_data_t;
static dbg_thread_data_t dbg_thread_data[DEBUG_MAX_THREADS];
#endif
static int dbg_nest_level = 0;
static char* INDENT_STR[] =
{
	"",
	"  ",
	"    ",
	"      ",
	"        ",
	"          ",
	"            "
};

/* Static process locale information */
static char ipp_lang[IPP_MAX_NAME] = { '\0' };

/**********************************************************************
 *  papi_init
 *********************************************************************/
void papi_init()
{
	char* dbg;
	int i;

	if (init_done)
		return;

	init_done = 1;
	dbg = getenv("DEBUG_PAPI");
	if (dbg == NULL)
		dbg = getenv("PAPI_DEBUG"); /* handle either one */
	if (dbg)
	{
		do_debug = atoi(dbg);
		if ( do_debug & DEBUG_MASK_FILE )
			dbgf = fopen("/tmp/papi.log", "w");
	}
}

/**********************************************************************
 *  papi_get_result
 *********************************************************************/
papi_status_t papi_get_result(papi_int_service_t* handle)
{
	papi_status_t rc = PAPI_OK;

	if (handle->response == NULL)
		rc = (papi_status_t)ippLastError();
	else
		rc = (papi_status_t)handle->response->request.status.status_code;

	return(rc);
}

/**********************************************************************
 *  papi_get_array_count
 *********************************************************************/
int papi_get_array_count(void** array)
{
	int i = 0;
	if (array)
		for(i=0; array[i] != NULL; i++);
	return(i);
}

/**********************************************************************
 *  papi_get_port_number
 *********************************************************************/
int papi_get_port_number(char* full_service_name)
{
	int port = 631;
	char method[HTTP_MAX_URI];
	char username[HTTP_MAX_URI];
	char host[HTTP_MAX_URI];
	char resource[HTTP_MAX_URI];
	
	httpSeparate(full_service_name, method, username, host, &port, resource);
	return(port);
}

/**********************************************************************
 *  papi_get_service_info
 *********************************************************************/
void papi_get_service_info(papi_int_service_t* handle,
		char* name, papi_object_type_t type)
{
	char* p_env;
	char method[HTTP_MAX_URI];
	char username[HTTP_MAX_URI];
	char resource[HTTP_MAX_URI];
	char* uri = "ipp://localhost:631";

	papi_dbg_printf(">papi_get_service_info\n");
	if (handle == NULL)
	{
		papi_dbg_printf("<papi_get_service_info returning from line %d\n", __LINE__);
		return;
	}

	(handle->full_service_name)[0] = '\0';
	(handle->object_uri)[0] = '\0';
	handle->port = 631;

    if (name == NULL)
    {
		papi_dbg_printf("<papi_get_service_info returning from line %d\n", __LINE__);
    	return;
    }

	if ( !strncmp(name, "http:", 5) || !strncmp(name, "ipp:", 4) )
	{
		/* A URI is specified so use it */
		uri = name;
		strcpy(handle->object_uri, name);
	}
	else
	{
		/* No URI specified, so default the full server name */
		p_env = getenv("IPPSERVER");
		if (p_env)
			uri = p_env;
	}
	
	/* Parse the URI */
	papi_dbg_printf("parsing uri: %s\n", uri);
	httpSeparate(uri, method, username, handle->hostname, &(handle->port), resource);

	/* Store the full service name */	
	sprintf(handle->full_service_name, "%s://%s:%d", method, handle->hostname, handle->port);
	papi_dbg_printf("handle->full_service_name: %s\n", handle->full_service_name);

	/* Store object URI if not already done */
	if ((handle->object_uri)[0] == '\0')
	{
		if (type == PAPI_PRINTER)
			sprintf(handle->object_uri, "%s/printers/%s", handle->full_service_name, name);
		else
			sprintf(handle->object_uri, "%s/jobs/%s", handle->full_service_name, name);
	}
	papi_dbg_printf("<papi_get_service_info returning OK\n");
}

/**********************************************************************
 *  papi_set_up_request
 *********************************************************************/
papi_status_t papi_set_up_request(papi_int_service_t* handle,
		char* name, papi_object_type_t type)
{
	papi_status_t rc;

	papi_init();
	papi_dbg_printf(">papi_set_up_request\n");

	if (handle == NULL)
	{
		papi_dbg_printf("<papi_set_up_request returning PAPI_BAD_REQUEST at line %d\n", __LINE__);
		return(PAPI_BAD_REQUEST);
	}

	/* Validate and connect to the server */
	rc = papi_validate_and_connect(INT_HANDLE(handle), name, type);
	if (rc != PAPI_OK)
	{
		papi_dbg_printf("<papi_set_up_request returning 0x%x at line %d\n", rc, __LINE__);
		return(rc);
	}
	if (INT_HANDLE(handle)->request == NULL)
	{
		papi_dbg_printf("<papi_set_up_request returning PAPI_INTERNAL_ERROR at line %d\n", __LINE__);
		return(PAPI_INTERNAL_ERROR);
	}

	/* Add common arguments */
	papi_add_common_args(INT_HANDLE(handle));

	papi_dbg_printf("<papi_set_up_request returning OK\n");
	return(PAPI_OK);
}

/**********************************************************************
 *  papi_set_language
 *********************************************************************/
void papi_set_language(papi_int_service_t* handle)
{
    char* loc = NULL;
    int i;
    int len;

	/* Get the current locale if not already done */
	if (ipp_lang[0] == '\0')
	{
		papi_dbg_printf(">papi_set_language\n");
		loc = setlocale(LC_ALL, "");
		if (loc)
		{
			/* Map the current locale to the IPP value */
			len = strlen(loc);
			if (len > 5)
				len = 5;
			for(i=0; i < len; i++)
			{
				if ((i == 2) && (loc[i] == '_'))
					ipp_lang[i] = '-';
				else
					ipp_lang[i] = tolower(loc[i]);
			}
			ipp_lang[len] = '\0';
			papi_dbg_printf("ipp_lang: %s\n", ipp_lang);
		}
		else
		{
			strcpy(ipp_lang, "en");
			papi_dbg_printf("ipp_lang defaulted to: en\n");
		}
		papi_dbg_printf("<papi_set_language\n");
	}

	/* Store it in the handle */
	strcpy(handle->language, ipp_lang);
}

/**********************************************************************
 *  papi_validate_and_connect
 *********************************************************************/
papi_status_t papi_validate_and_connect(papi_int_service_t* handle,
		char* name, papi_object_type_t type)
{
	http_t* http = NULL;

	/* Ensure global initialization has been done */
	papi_init();
	papi_dbg_printf(">papi_validate_and_connect\n");

	/* Validate arguments */
	if (handle == NULL)
		return(PAPI_BAD_REQUEST);

	/* Get rid of old connection and request (if any) */
	papi_clear_prev_request(handle);

	/* Get the full service name URI */
	papi_get_service_info(handle, name, type);
	if (handle->full_service_name == NULL)
	{
		papi_dbg_printf("<papi_validate_and_connect returning PAPI_BAD_REQUEST from line %d\n", __LINE__);
		return(PAPI_BAD_REQUEST);
	}
	
	/* Connect to the host */
	papi_dbg_printf("  handle->hostname: %s\n", handle->hostname);
	papi_dbg_printf("      handle->port: %d\n", handle->port);
	papi_dbg_printf("handle->encryption: %d\n", handle->encryption);
	handle->http = httpConnectEncrypt(handle->hostname,
						handle->port,
						handle->encryption);
	if (handle->http == NULL)
	{
		papi_dbg_printf("<papi_validate_and_connect returning PAPI_SERVICE_UNAVAILABLE from line %d\n", __LINE__);
		return(PAPI_SERVICE_UNAVAILABLE);
	}

	/* Create a new request */
	handle->request = ippNew();
	if (handle->request == NULL)
	{
		httpClose(handle->http);
		papi_dbg_printf("<papi_validate_and_connect returning PAPI_SERVICE_UNAVAILABLE from line %d\n", __LINE__);
		return(PAPI_SERVICE_UNAVAILABLE);
	}

	/* Initialize other fields */
	papi_set_language(handle);
	strcpy(handle->char_set, "utf8");

	papi_dbg_printf("<papi_validate_and_connect returning OK\n");
	return(PAPI_OK);
}

/**********************************************************************
 *  papi_clear_prev_request
 *********************************************************************/
void papi_clear_prev_request(papi_int_service_t* handle)
{
	if (handle->status)
	{
		papiAttributeListFree(handle->status);
		handle->status = NULL;
	}
	if (handle->capabilities)
	{
		papiAttributeListFree(handle->capabilities);
		handle->capabilities = NULL;
	}
	if (handle->response)
	{
		ippDelete(handle->response);
		handle->response = NULL;
	}
	if (handle->http)
	{
		httpClose(handle->http);
		handle->http = NULL;
	}
}

/**********************************************************************
 *  papi_add_common_args
 *********************************************************************/
void papi_add_common_args(papi_int_service_t* handle)
{
	if (handle == NULL)
		return;
	if (handle->request == NULL)
		return;

	if (handle->char_set)
		ippAddString(handle->request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,	
				"attributes-charset", NULL, handle->char_set);
	if (handle->language)
		ippAddString(handle->request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,	
				"attributes-natural-language", NULL, handle->language);
	if (handle->user_name)
	    ippAddString(handle->request, IPP_TAG_OPERATION, IPP_TAG_NAME,
				"requesting-user-name", NULL, handle->user_name);
}

/**********************************************************************
 *  papi_do_request
 *********************************************************************/
void papi_do_request(papi_int_service_t* handle, char* resource,
							char** files)
{
	ipp_t* response = NULL;
	ipp_t* saved_req = NULL;
	ipp_attribute_t* job_uri = NULL;
	ipp_attribute_t* job_id = NULL;
	ipp_attribute_t* last_doc = NULL;
	int i;

	papi_dbg_printf(">papi_do_request\n");
	if ((files != NULL) && (files[0] != NULL))
	{
		if (files[1] == NULL)
		{
			/*  Single-document request... */
		 	handle->response = ippDoFileRequest(handle->http,
										    handle->request,
										    resource,
											files[0]);
		}
		else
		{
			papi_dbg_printf("multi-document request...\n");

			/*  First, create the job */
			handle->request->request.op.operation_id = IPP_CREATE_JOB;
			saved_req = papi_dup_cups_request(handle->request);
		 	handle->response = ippDoRequest(handle->http,
										    handle->request,
										    resource);
			if (handle->response == NULL)
			{
				papi_dbg_printf("ippDoRequest(IPP_CREATE_JOB) failed!\n");
				papi_dbg_printf("<papi_do_request\n");
				return;
			}
				
			/* Put the job-uri into the request for the
			// subsequent send-document requests */
			job_uri = ippFindAttribute(handle->response, "job-uri", IPP_TAG_URI);
			if (job_uri == NULL)
			{
				papi_dbg_printf("Could not find job-uri attribute\n");
				papi_dbg_printf("<papi_do_request\n");
				return;
			}			
			ippAddString(saved_req, IPP_TAG_OPERATION, IPP_TAG_URI,
				"job-uri", NULL, job_uri->values[0].string.text);
				
			/* Put the job-id into the request for the
			// subsequent send-document requests */
			job_id = ippFindAttribute(handle->response, "job-id", IPP_TAG_INTEGER);
			if (job_id == NULL)
			{
				papi_dbg_printf("Could not find job-id attribute\n");
				papi_dbg_printf("<papi_do_request\n");
				return;
			}			
			ippAddInteger(saved_req, IPP_TAG_OPERATION, IPP_TAG_INTEGER,
				"job-id", job_id->values[0].integer);

			ippDelete(handle->response);

			/* Do the send-document for each document in the job */
			saved_req->request.op.operation_id = IPP_SEND_DOCUMENT;
			ippAddBoolean(saved_req, IPP_TAG_OPERATION, "last-document", 0);
			for(i=0; (files[i] != NULL) && (handle->response != NULL); i++)
			{
				papi_dbg_printf("sending: %s\n", files[i]);
				handle->request = papi_dup_cups_request(saved_req);
				if (files[i+1] == NULL)
				{
					/* Last document */
					last_doc = ippFindAttribute(handle->request, "last-document", IPP_TAG_BOOLEAN);
					if (last_doc)
						last_doc->values[0].boolean = 1; /* Last document */
				}
			 	handle->response = ippDoFileRequest(handle->http,
										    handle->request,
										    resource,
											files[i]);
				if ((handle->response != NULL) && (files[i+1] != NULL))
					ippDelete(handle->response);
			}
			ippDelete(saved_req);
		}
	}
	else
	{
	 	handle->response = ippDoRequest(handle->http,
									    handle->request,
									    resource);
	}
	papi_dbg_printf("<papi_do_request\n");
}

/**********************************************************************
 *  papi_dbg_printf
 *********************************************************************/
void papi_dbg_printf(char* fmt, ...)
{
	char* actual_fmt;
	char* indent;
	va_list arg_ptr;

	if (do_debug)
	{
		/* Set up for proper indenting */
		actual_fmt = malloc( strlen(fmt) + 20 );
		if ((fmt[0] == '<') && (dbg_nest_level > 0))
			dbg_nest_level--;
		if (dbg_nest_level < 7)
			indent = INDENT_STR[dbg_nest_level];
		else
			indent = INDENT_STR[6];
		sprintf(actual_fmt, "%s%s", indent, fmt);

		/* Trace the data */
		if (do_debug & DEBUG_MASK_STDOUT)
		{
			va_start(arg_ptr, fmt);				
			vprintf(actual_fmt, arg_ptr);
			va_end(arg_ptr);
			fflush(stdout);
		}
		if ((do_debug & DEBUG_MASK_FILE) && (dbgf != NULL))
		{
			va_start(arg_ptr, fmt);				
			vfprintf(dbgf, actual_fmt, arg_ptr);
			va_end(arg_ptr);
			fflush(dbgf);
		}

		free(actual_fmt);

		if (fmt[0] == '>')
			dbg_nest_level++;
	}
}

/**********************************************************************
 *  papi_dbg_list_cups_attribute
 *********************************************************************/
void papi_dbg_list_cups_attribute(ipp_attribute_t* attr)
{
	int i;
	int yr;

	if (!do_debug)
		return;
	if (attr == NULL)
		return;

	papi_dbg_printf("   %s ", attr->name);
	for(i=0; i<attr->num_values; i++)
	{
		if (i>0) printf(", ");
		switch(attr->value_tag)
		{
		case IPP_TAG_STRING:
			papi_dbg_printf("(IPP_TAG_STRING): ");
			papi_dbg_printf("%s", attr->values[i].string.text);
			break;
		case IPP_TAG_NAME:
			papi_dbg_printf("(IPP_TAG_NAME): ");
			papi_dbg_printf("%s", attr->values[i].string.text);
			break;
		case IPP_TAG_TEXT:
			papi_dbg_printf("(IPP_TAG_TEXT): ");
			papi_dbg_printf("%s", attr->values[i].string.text);
			break;
		case IPP_TAG_URI:
			papi_dbg_printf("(IPP_TAG_URI): ");
			papi_dbg_printf("%s", attr->values[i].string.text);
			break;
		case IPP_TAG_CHARSET:
			papi_dbg_printf("(IPP_TAG_CHARSET): ");
			papi_dbg_printf("%s", attr->values[i].string.text);
			break;
		case IPP_TAG_KEYWORD:
			papi_dbg_printf("(IPP_TAG_KEYWORD): ");
			papi_dbg_printf("%s", attr->values[i].string.text);
			break;
		case IPP_TAG_LANGUAGE:
			papi_dbg_printf("(IPP_TAG_LANGUAGE): ");
			papi_dbg_printf("%s", attr->values[i].string.text);
			break;
		case IPP_TAG_MIMETYPE:
			papi_dbg_printf("(IPP_TAG_MIMETYPE): ");
			papi_dbg_printf("%s", attr->values[i].string.text);
			break;
		case IPP_TAG_BOOLEAN:
			papi_dbg_printf("(IPP_TAG_BOOLEAN): ");
			if (attr->values[i].boolean)
				papi_dbg_printf("TRUE");
			else
				papi_dbg_printf("FALSE");
        	break;
		case IPP_TAG_INTEGER:
			papi_dbg_printf("(IPP_TAG_INTEGER): ");
			papi_dbg_printf("%d", attr->values[i].integer);
        	break;
		case IPP_TAG_ENUM:
			papi_dbg_printf("(IPP_TAG_ENUM): ");
			papi_dbg_printf("%d", attr->values[i].integer);
        	break;
		case IPP_TAG_RANGE:
			papi_dbg_printf("(IPP_TAG_RANGE): ");
			papi_dbg_printf("%d to %d", attr->values[i].range.lower, attr->values[i].range.upper);
        	break;
		case IPP_TAG_DATE:
			papi_dbg_printf("(IPP_TAG_DATE): ");
			yr = attr->values[i].date[1] + (attr->values[i].date[0]*0x100);
			papi_dbg_printf("%02d/%02d/%04d %02d:%02d:%02d UTC",
				attr->values[i].date[2],
				attr->values[i].date[3],
				yr,
				attr->values[i].date[4],
				attr->values[i].date[5],
				attr->values[i].date[6]);
			break;
		default:
			papi_dbg_printf("unrecognized value_tag=0x%x", attr->value_tag);
			break;
		}
	}
	papi_dbg_printf("\n");
}

/**********************************************************************
 *  papi_dbg_list_all_cups_attributes
 *********************************************************************/
void papi_dbg_list_all_cups_attributes(ipp_t* response, char* str)
{
	ipp_attribute_t* attr;

	if (!do_debug)
		return;

	if (str)
		papi_dbg_printf(str);
	if (response == NULL)
	{
		papi_dbg_printf("   No attributes present\n");
		return;
	}

	for( attr=response->attrs; attr != NULL; attr=attr->next )
		papi_dbg_list_cups_attribute(attr);
}

/**********************************************************************
 *  papi_dbg_list_some_cups_attributes
 *********************************************************************/
void papi_dbg_list_some_cups_attributes(ipp_t* response, char* attrs[],
	int num_attrs, char* str)
{
	ipp_attribute_t* attr;
	int i;

	if (!do_debug)
		return;
	if (str)
		printf(str);
	if (response == NULL)
	{
		papi_dbg_printf("   No attributes present\n");
		return;
	}

	for(i=0; i<num_attrs; i++)
	{
	    attr=ippFindAttribute(response, attrs[i], IPP_TAG_ZERO);
		papi_dbg_list_cups_attribute(attr);
	}
}

