/**********************************************************************
 *
 * Open Standard Print API (PAPI) - Base High-level Functions
 *
 * Copyright (c) 2002 International Business Machines Corp.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "papi.h"
#include "papi_int.h"

/*---------------------------------------------------------------------
 * Internal functions
 *-------------------------------------------------------------------*/

typedef enum
{
	SUB_JOB_NORMAL = 0,
	SUB_JOB_VALIDATE_ONLY,
	SUB_JOB_SETUP_ONLY
} sub_job_type_t;

#define JT_CHUNK_SIZE 1023

/**********************************************************************
 *  papi_sub_add_jt_attr
 *********************************************************************/
papi_status_t papi_sub_add_jt_attr(
                papi_int_service_t*   handle,
          char*                 jt_buffer )
{
	papi_status_t rc = PAPI_OK;
	int i;
	char save_ch;
	char* curr_buff = NULL;
	unsigned int num_bytes;
	unsigned int bytes_left;
	int num_values = 0;
	ipp_attribute_t* jt_attr;
	ipp_value_t* jt_value;

	/* Initialize pointers and counters */
	curr_buff = jt_buffer;
	bytes_left = strlen(jt_buffer);
	num_values = (bytes_left / JT_CHUNK_SIZE) + 1;

	/* Create the job ticket attribute and init it */
	jt_attr = papi_add_cups_attr(handle->request, num_values);

	jt_attr->name = strdup("job-ticket");
	jt_attr->group_tag = IPP_TAG_JOB;
	jt_attr->value_tag = IPP_TAG_TEXT;

	/* Add all the buffers of data as attr values */
	for(i=0, jt_value = jt_attr->values;
	    i < num_values;
		i++, jt_value++)
	{

		if (bytes_left > JT_CHUNK_SIZE)	
		{
			save_ch = curr_buff[JT_CHUNK_SIZE];
			curr_buff[JT_CHUNK_SIZE] = '\0';
			jt_value->string.charset = NULL;
			jt_value->string.text = strdup(curr_buff);
			curr_buff[JT_CHUNK_SIZE] = save_ch;
			bytes_left -= JT_CHUNK_SIZE;
			curr_buff += JT_CHUNK_SIZE;
		}
		else
		{
			if (bytes_left > 0)
			{
				jt_value->string.charset = NULL;
				jt_value->string.text = strdup(curr_buff);
			}
		}
	}

	return(rc);
}

/**********************************************************************
 *  papi_sub_add_job_ticket
 *********************************************************************/
papi_status_t papi_sub_add_job_ticket(
                papi_int_service_t*   handle,
          papi_job_ticket_t*    job_ticket )
{
	papi_status_t rc;
	int i;
	char* jt_buffer = NULL;
	char* jt_filename = NULL;
	char save_ch;
	int en;
	struct stat st;
	FILE* jtf = NULL;

	papi_dbg_printf(">papi_sub_add_job_ticket\n");
	if (job_ticket == NULL)
	{
		papi_dbg_printf("<papi_sub_add_job_ticket returning OK (no JT)\n");
		return(PAPI_OK);
	}

	papi_dbg_printf("job ticket specified\n");
	/*-----------------------------------------------------------------
	 * Get all the job ticket data into a buffer (we currently
	 * will not support passing the job ticket by reference to
	 * the server).
	 *---------------------------------------------------------------*/
	if (job_ticket->ticket_data != NULL)
	{
		/* We're all set! */
		jt_buffer = job_ticket->ticket_data;
	}
	else
	{
		/* Get a pointer to the filename */
		if  (job_ticket->file_name != NULL)
		{
			jt_filename = (char*)job_ticket->file_name;
			if (job_ticket->file_name[0] != DIR_DELIM_CH)
			{
				if (strlen(job_ticket->file_name) > 9)
				{
					save_ch = jt_filename[8];
					jt_filename[8] = '\0';
					if ( !strcmp(jt_filename, "file:///") )
					{
						jt_filename[8] = save_ch;
						jt_filename = &(jt_filename[7]);
					}
					else
					{
						jt_filename[8] = save_ch;
					}
				}
			}
		}
		else
		{
			papi_dbg_printf("no job ticket data or filename specified\n");
			papi_dbg_printf("<papi_sub_add_job_ticket returning PAPI_BAD_REQUEST\n");
			return(PAPI_BAD_REQUEST);
		}

		/* Allocate the buffer and read in the job ticket */
		if (stat(jt_filename, &st) < 0)
		{
			en = errno;
			papi_dbg_printf("stat(%s) failed. errno=%d\n", jt_filename, en);
			papi_dbg_printf("<papi_sub_add_job_ticket returning PAPI_BAD_REQUEST\n");
			return(PAPI_BAD_REQUEST);
		}
		jt_buffer = malloc( st.st_size+1 );
		jtf = fopen(jt_filename, "rb");
		if (jtf == NULL)
		{
			en = errno;
			free(jt_buffer);
			papi_dbg_printf("fopen(%s) failed. errno=%d\n", jt_filename, en);
			papi_dbg_printf("<papi_sub_add_job_ticket returning PAPI_BAD_REQUEST\n");
			return(PAPI_BAD_REQUEST);
		}
		if (fread(jt_buffer, 1, st.st_size, jtf) != st.st_size)
		{
			en = errno;
			free(jt_buffer);
			fclose(jtf);
			papi_dbg_printf("fread(%s) failed. errno=%d\n", jt_filename, en);
			papi_dbg_printf("<papi_sub_add_job_ticket returning PAPI_BAD_REQUEST\n");
			return(PAPI_BAD_REQUEST);
		}
		fclose(jtf);
		jt_buffer[st.st_size] = '\0';
	} /* end "filename of JT passed" */

	/*-----------------------------------------------------------------
	 * Add the job ticket attribute to the job request
	 *---------------------------------------------------------------*/
	rc = papi_sub_add_jt_attr(handle, jt_buffer);

	/*-----------------------------------------------------------------
	 * Clean up and return
	 *---------------------------------------------------------------*/
	if (job_ticket->ticket_data == NULL)
		free(jt_buffer);
	papi_dbg_printf("<papi_sub_add_job_ticket returning %d\n", rc);
	return(rc);
}

/**********************************************************************
 *  papi_job_ticket_cups2jt
 *********************************************************************/
void papi_job_ticket_cups2jt(ipp_attribute_t* jt_attr, papi_job_ticket_t* jt)
{
	int i;
	unsigned long buffsize = 0;
	ipp_value_t* jt_value;

	/* Initialize the fields */
	jt->format = PAPI_JT_FORMAT_JDF;
	jt->ticket_data = NULL;
	jt->file_name = NULL;

	if ( !strcmp(jt_attr->name, "job-ticket-reference") )
	{
		/* Handle job-ticket-reference... */
		jt->file_name = strdup(jt_attr->values[0].string.text);
	}
	else if ( !strcmp(jt_attr->name, "job-ticket") )
	{
		/* Handle job-ticket... */

		/* Compute the size of buffer needed */
		for(i=0, jt_value = jt_attr->values;
		    i < jt_attr->num_values;
			i++, jt_value++)
		{
			buffsize += strlen(jt_value->string.text);		
		}
		buffsize += 1;

		/* Allocate the buffer and copy the data */
		jt->ticket_data = malloc(buffsize);
		jt->ticket_data[0] = '\0';
		for(i=0, jt_value = jt_attr->values;
		    i < jt_attr->num_values;
			i++, jt_value++)
		{
			strcat(jt->ticket_data, jt_value->string.text);
		}
	}
}

/**********************************************************************
 *  papi_sub_job
 *********************************************************************/
papi_status_t papi_sub_job(
                papi_service_t        handle,
          char*                 printer,
          papi_attribute_t**    job_attributes,
          papi_job_ticket_t*    job_ticket,
          char**                file_names,
                papi_job_t*           job,
          sub_job_type_t		  sub_type)
{
	papi_status_t rc;
	int i;
	papi_attribute_t* attr = NULL;
	papi_int_job_t* int_job = NULL;

	papi_init();
	papi_dbg_printf(">papi_sub_job\n");

	/* Set up the request */
	rc = papi_set_up_request(INT_HANDLE(handle), printer, PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papi_sub_job", handle, rc);

	/* Add operation-specific data to the request */
	ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_URI,	
				"printer-uri", NULL, INT_HANDLE(handle)->object_uri);

	/* First add the operation group attributes */
	rc = papi_attr_papi2cups(job_attributes,
							INT_HANDLE(handle)->request,
							NULL,
							IPP_TAG_OPERATION);
	if (rc != PAPI_OK)
		PAPI_RETURN("papi_sub_job", handle, rc);

	/* Ensure we have a reasonable job-name set */
	attr = papiAttributeListFind(job_attributes, "job-name");
	if ((attr == NULL) && (file_names != NULL) && (file_names[0] != NULL) &&
	    (job_ticket == NULL))
	{
		ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_NAME,	
					"job-name", NULL, file_names[0]);
	}

	/* Next add the job template group attributes */
	rc = papi_attr_papi2cups(job_attributes,
							INT_HANDLE(handle)->request,
							NULL,
							IPP_TAG_JOB);
	if (rc != PAPI_OK)
		PAPI_RETURN("papi_sub_job", handle, rc);

	papi_dbg_list_all_cups_attributes(INT_HANDLE(handle)->request, "   Job attributes:\n");

	/* Add the job ticket (if any) */
	rc = papi_sub_add_job_ticket(INT_HANDLE(handle), job_ticket);
	if (rc != PAPI_OK)
		PAPI_RETURN("papi_sub_job", handle, rc);

	/* Set up the proper request and do it  */
	INT_HANDLE(handle)->request->request.any.request_id = 1;
	if ( sub_type == SUB_JOB_VALIDATE_ONLY )
	{
		INT_HANDLE(handle)->request->request.op.operation_id = IPP_VALIDATE_JOB;
	 	papi_do_request(handle, "/printers/", NULL);
	}
	else if ( sub_type == SUB_JOB_SETUP_ONLY )
	{
		INT_HANDLE(handle)->request->request.op.operation_id = IPP_PRINT_JOB;
		papi_dbg_printf("<papi_sub_job (setup-only) returning 0x%x\n", rc);
		return(rc);
	}
	else
	{
		INT_HANDLE(handle)->request->request.op.operation_id = IPP_PRINT_JOB;
	 	papi_do_request(handle, "/printers/", file_names);
	}
	if (INT_HANDLE(handle)->response == NULL)
		PAPI_RETURN("papi_sub_job", handle, PAPI_BAD_REQUEST);

	rc = papi_get_result(INT_HANDLE(handle));

	/* Create the job object to be returned */
	int_job = malloc(sizeof(papi_int_job_t));
	if (int_job == NULL)
		PAPI_RETURN("papi_sub_job", handle, PAPI_INTERNAL_ERROR);
	int_job->name = strdup(printer);
	int_job->id = -1;
	int_job->job_ticket = NULL;
	int_job->attributes = NULL;
	*job = (papi_job_t)int_job;

	/* Convert returned attributes to PAPI object/attributes */
	if (rc == PAPI_OK)
	{
		rc = papi_attr_cups2papi(INT_HANDLE(handle)->response->attrs,
                           &(int_job->attributes));
	    attr = papiAttributeListFind(int_job->attributes, "job-id");
	    if (attr && attr->values[0])
	    	int_job->id = attr->values[0]->integer;
	}

	PAPI_RETURN("papi_sub_job", handle, rc);
}

/**********************************************************************
 * papi_do_job_op
 *********************************************************************/
papi_status_t papi_do_job_op(
                papi_service_t      handle,
		  ipp_op_t			op,
          char*               printer,
          int32_t             job_id,
          char*               hold_until,
          time_t*             hold_until_time )
{
	papi_status_t rc;
	int i;
	papi_attribute_t* attr = NULL;
	papi_int_job_t* int_job = NULL;

	papi_init();
	papi_dbg_printf(">papi_do_job_op(0x%x)\n", op);

	/* Set up the request */
	rc = papi_set_up_request(INT_HANDLE(handle), printer, PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papi_do_job_op", handle, rc);

	/* Add operation-specific data to the request */
	ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_URI,	
				"printer-uri", NULL, INT_HANDLE(handle)->object_uri);
	ippAddInteger(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,	
					"job-id", job_id);

	if (op == IPP_HOLD_JOB)
	{
		if (hold_until)
		{
			if ( !strcmp(hold_until, "timed") )
			{
				httpClose(INT_HANDLE(handle)->http);
				INT_HANDLE(handle)->http = NULL;
				PAPI_RETURN("papi_do_job_op", handle, PAPI_BAD_REQUEST);
			}
			ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,	
					"job-hold-until", NULL, hold_until);
		}
	}

	papi_dbg_list_all_cups_attributes(INT_HANDLE(handle)->request, "   Request attributes:\n");

	/* Set up the request and do it  */
	INT_HANDLE(handle)->request->request.any.request_id = 1;
	INT_HANDLE(handle)->request->request.op.operation_id = op;
 	papi_do_request(handle, "/printers/", NULL);

	rc = papi_get_result(INT_HANDLE(handle));
	PAPI_RETURN("papi_do_job_op", handle, rc);
}

/*---------------------------------------------------------------------
 * Job API
 *-------------------------------------------------------------------*/

/**********************************************************************
 * papiJobSubmit
 *********************************************************************/
papi_status_t papiJobSubmit(
                papi_service_t        handle,
          char*                 printer,
          papi_attribute_t**    job_attrs,
          papi_job_ticket_t*    job_ticket,
          char**                file_names,
                papi_job_t*           job )
{
	return( papi_sub_job(handle, printer, job_attrs, job_ticket, file_names, job, SUB_JOB_NORMAL) );
}

/**********************************************************************
 * papiJobSubmitByReference
 *********************************************************************/
papi_status_t papiJobSubmitByReference(
                papi_service_t        handle,
          char*                 printer,
          papi_attribute_t**    job_attrs,
          papi_job_ticket_t*    job_ticket,
          char**                file_names,
                papi_job_t*           job )
{
	/* ???????? for now, just handle it the same as papiJobSubmit ?????? */
	return( papi_sub_job(handle, printer, job_attrs, job_ticket, file_names, job, SUB_JOB_NORMAL) );
}

/**********************************************************************
 * papiJobValidate
 *********************************************************************/
papi_status_t papiJobValidate(
                papi_service_t        handle,
          char*                 printer,
          papi_attribute_t**    job_attrs,
          papi_job_ticket_t*    job_ticket,
          char**                file_names,
                papi_job_t*           job )
{
	return( papi_sub_job(handle, printer, job_attrs, job_ticket, file_names, job, SUB_JOB_VALIDATE_ONLY) );
}

/**********************************************************************
 * papiJobStreamOpen
 *********************************************************************/
papi_status_t papiJobStreamOpen(
                papi_service_t       handle,
          char*                printer,
          papi_attribute_t**   job_attrs,
          papi_job_ticket_t*   job_ticket,
                papi_stream_t*       stream )
{
	papi_status_t rc;
	papi_int_stream_t* int_stream = NULL;
	int en;

	papi_dbg_printf(">papiJobStreamOpen\n");
	/* Create the stream structure and open the temp file for print data */
	int_stream = malloc( sizeof(papi_int_stream_t) );
	*stream = (papi_stream_t)int_stream;
	strcpy(int_stream->printer, printer);
	tmpnam(int_stream->temp_filename);
	if (int_stream->temp_filename[0] == '\0')
	{
		papi_dbg_printf("tmpnam returned an empty name\n");
		PAPI_RETURN("papiJobStreamOpen", handle, PAPI_INTERNAL_ERROR);
	}
	int_stream->tempf = fopen( int_stream->temp_filename, "wb" );
	if ( int_stream->tempf == NULL )
	{
		en = errno;
		papi_dbg_printf("fopen(%s) failed. errno=%d\n", int_stream->temp_filename, en);
		PAPI_RETURN("papiJobStreamOpen", handle, PAPI_INTERNAL_ERROR);
	}
	papi_dbg_printf("fopen(%s) successful\n", int_stream->temp_filename);
	
	/* Set up the print job */
	rc = papi_sub_job(handle, printer, job_attrs, job_ticket, NULL, NULL, SUB_JOB_SETUP_ONLY);
	papi_dbg_printf("<papiJobStreamOpen returning 0x%x\n", rc);
	return(rc);
}

/**********************************************************************
 * papiJobStreamWrite
 *********************************************************************/
papi_status_t papiJobStreamWrite(
                papi_service_t       handle,
                papi_stream_t        stream,
          void*                buffer,
          size_t               buflen )
{
	int byte_count;
	if ((stream == NULL) || (INT_STREAM(stream)->tempf == NULL) ||
        (buffer == NULL) || (buflen < 1))
		return(PAPI_BAD_REQUEST);

	byte_count = fwrite(buffer, 1, buflen, INT_STREAM(stream)->tempf);
	if (byte_count != buflen)
		return(PAPI_INTERNAL_ERROR);

	return(PAPI_OK);
}

/**********************************************************************
 * papiJobStreamClose
 *********************************************************************/
papi_status_t papiJobStreamClose(
                papi_service_t       handle,
                papi_stream_t        stream,
                papi_job_t*          job )
{
	papi_status_t rc;
	int i;
	papi_attribute_t* attr = NULL;
	papi_int_job_t* int_job = NULL;
	char* files[2] = { NULL, NULL };

	papi_dbg_printf(">papiJobStreamClose\n");

	if ((stream == NULL) || (INT_STREAM(stream)->tempf == NULL))
		PAPI_RETURN("papiJobStreamClose", handle, PAPI_OK);

	/* Close the temporary file */
	fclose(INT_STREAM(stream)->tempf);

	/* Submit the job */
 	files[0] = INT_STREAM(stream)->temp_filename;
 	papi_do_request(INT_HANDLE(handle), "/printers/", files);
	unlink(INT_STREAM(stream)->temp_filename);
	if (INT_HANDLE(handle)->response == NULL)
		PAPI_RETURN("papiJobStreamClose", handle, PAPI_BAD_REQUEST);

	rc = papi_get_result(INT_HANDLE(handle));

	/* Create the job object to be returned */
	int_job = malloc(sizeof(papi_int_job_t));
	if (int_job == NULL)
		PAPI_RETURN("papiJobStreamClose", handle, PAPI_INTERNAL_ERROR);
	int_job->name = strdup(INT_STREAM(stream)->printer);
	int_job->id = -1;
	int_job->job_ticket = NULL;
	int_job->attributes = NULL;
	*job = (papi_job_t)int_job;

	/* Convert returned attributes to PAPI object/attributes */
	if (rc == PAPI_OK)
	{
		rc = papi_attr_cups2papi(INT_HANDLE(handle)->response->attrs,
                           &(int_job->attributes));
	    attr = papiAttributeListFind(int_job->attributes, "job-id");
	    if (attr && attr->values[0])
	    	int_job->id = attr->values[0]->integer;
	}

	PAPI_RETURN("papiJobStreamClose", handle, rc);
}

/**********************************************************************
 * papiJobQuery
 *********************************************************************/
papi_status_t papiJobQuery(
                papi_service_t     handle,
          char*               printer_name,
          int32_t             job_id,
          char*               requested_attrs[],
                papi_job_t*         job )
{
	papi_status_t rc;
	int i;
	papi_attribute_t* attr = NULL;
	papi_int_job_t* int_job = NULL;

	papi_init();
	papi_dbg_printf(">papiJobQuery\n");

	if (printer_name == NULL)
	{
		papi_dbg_printf("<papiJobQuery returning PAPI_BAD_REQUEST at line %d\n", __LINE__);
		return(PAPI_BAD_REQUEST);
	}
	papi_dbg_printf("querying printer %s for job-id %d\n", printer_name, job_id);

	/* Set up the request */
	rc = papi_set_up_request(INT_HANDLE(handle), printer_name, PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papiJobQuery", handle, rc);

	/* Add operation-specific data to the request */
	ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_URI,	
				"printer-uri", NULL, INT_HANDLE(handle)->object_uri);
	ippAddInteger(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,	
					"job-id", job_id);

	if ((requested_attrs != NULL) && (requested_attrs[0] != NULL))
	{
		/* set requested attributes */
		for(i=0; requested_attrs[i] != NULL; i++) ;
		ippAddStrings(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_NAME,
	                "requested-attributes",
					i,
					NULL, requested_attrs);
	}
	papi_dbg_list_all_cups_attributes(INT_HANDLE(handle)->request, "   Request attributes:\n");

	INT_HANDLE(handle)->request->request.op.operation_id = IPP_GET_JOB_ATTRIBUTES;
	INT_HANDLE(handle)->request->request.any.request_id = 1;

	/* Do the request */
 	papi_do_request(INT_HANDLE(handle), "/printers", NULL);
	if (INT_HANDLE(handle)->response == NULL)
		PAPI_RETURN("papiJobQuery", handle, PAPI_BAD_REQUEST);

	rc = papi_get_result(INT_HANDLE(handle));

	/* Create the job object to be returned */
	int_job = malloc(sizeof(papi_int_job_t));
	if (int_job == NULL)
		PAPI_RETURN("papiJobQuery", handle, PAPI_INTERNAL_ERROR);
	int_job->name = strdup(printer_name);
	int_job->id = job_id;
	int_job->job_ticket = NULL;
	int_job->attributes = NULL;
	*job = (papi_job_t)int_job;

	if (rc == PAPI_OK)
	{
		/* Convert returned attributes to PAPI object/attributes */
		rc = papi_attr_cups2papi(INT_HANDLE(handle)->response->attrs,
	                           &(int_job->attributes));
	}

	PAPI_RETURN("papiJobQuery", handle, rc);
}

/**********************************************************************
 * papiJobModify
 *********************************************************************/
papi_status_t papiJobModify(
                papi_service_t     handle,
          char*               printer_name,
          int32_t             job_id,
          papi_attribute_t**  job_attrs,
                papi_job_t*         job )
{
	papi_status_t rc;
	int i;
	papi_attribute_t* attr = NULL;
	papi_int_job_t* int_job = NULL;

	papi_init();
	papi_dbg_printf(">papiJobModify\n");

	if ((printer_name == NULL) || (job_attrs == NULL))
	{
		papi_dbg_printf("<papiJobModify returning PAPI_BAD_REQUEST at line %d\n", __LINE__);
		return(PAPI_BAD_REQUEST);
	}
	papi_dbg_printf("modifying printer %s job-id %d\n", printer_name, job_id);

	/* Set up the request */
	rc = papi_set_up_request(INT_HANDLE(handle), printer_name, PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papiJobModify", handle, rc);

	/* Add operation-specific data to the request */
	ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_URI,	
				"printer-uri", NULL, INT_HANDLE(handle)->object_uri);
	ippAddInteger(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,	
					"job-id", job_id);

	/* Add the set attributes */
	rc = papi_attr_papi2cups(job_attrs,
                             INT_HANDLE(handle)->request,
                             NULL,
                             IPP_TAG_JOB);
    if (rc != PAPI_OK)
		PAPI_RETURN("papiJobModify", handle, rc);

	papi_dbg_list_all_cups_attributes(INT_HANDLE(handle)->request, "   Request attributes:\n");

	INT_HANDLE(handle)->request->request.op.operation_id = IPP_SET_JOB_ATTRIBUTES;
	INT_HANDLE(handle)->request->request.any.request_id = 1;

	/* Do the request */
 	papi_do_request(INT_HANDLE(handle), "/printers", NULL);
	if (INT_HANDLE(handle)->response == NULL)
		PAPI_RETURN("papiJobModify", handle, PAPI_BAD_REQUEST);

	rc = papi_get_result(INT_HANDLE(handle));

	/* Create the job object to be returned */
	int_job = malloc(sizeof(papi_int_job_t));
	if (int_job == NULL)
		PAPI_RETURN("papiJobModify", handle, PAPI_INTERNAL_ERROR);
	int_job->name = strdup(printer_name);
	int_job->id = job_id;
	int_job->job_ticket = NULL;
	int_job->attributes = NULL;
	*job = (papi_job_t)int_job;

	if (rc == PAPI_OK)
	{
		/* Convert returned attributes to PAPI object/attributes */
		rc = papi_attr_cups2papi(INT_HANDLE(handle)->response->attrs,
	                           &(int_job->attributes));
	}

	PAPI_RETURN("papiJobModify", handle, rc);
}

/**********************************************************************
 * papiJobCancel
 *********************************************************************/
papi_status_t papiJobCancel(
                papi_service_t     handle,
          char*               printer,
          int32_t             job_id )
{
	return( papi_do_job_op(handle, IPP_CANCEL_JOB, printer, job_id, NULL, NULL ) );
}

/**********************************************************************
 * papiJobHold
 *********************************************************************/
papi_status_t papiJobHold(
                papi_service_t      handle,
          char*               printer,
          int32_t             job_id,
          char*               hold_until,
          time_t*             hold_until_time )
{
	return( papi_do_job_op(handle, IPP_HOLD_JOB, printer, job_id, hold_until, hold_until_time ) );
}

/**********************************************************************
 * papiJobRelease
 *********************************************************************/
papi_status_t papiJobRelease(
                papi_service_t      handle,
          char*               printer,
          int32_t             job_id )
{
	return( papi_do_job_op(handle, IPP_RELEASE_JOB, printer, job_id, NULL, NULL ) );
}

/**********************************************************************
 * papiJobRestart
 *********************************************************************/
papi_status_t papiJobRestart(
                papi_service_t      handle,
          char*               printer,
          int32_t             job_id )
{
	return( papi_do_job_op(handle, IPP_RESTART_JOB, printer, job_id, NULL, NULL ) );
}

/**********************************************************************
 * papiJobGetAttributeList
 *********************************************************************/
papi_attribute_t** papiJobGetAttributeList(papi_job_t job)
{
	if (job)
		return( INT_JOB_HANDLE(job)->attributes );
	else
		return(NULL);
}

/**********************************************************************
 * papiJobGetPrinterName
 *********************************************************************/
char* papiJobGetPrinterName(papi_job_t job)
{
	papi_int_job_t *j = INT_JOB_HANDLE(job);
	char *result = NULL;

	if (job != NULL) {
		result = INT_JOB_HANDLE(job)->name;
		if ((result == NULL) || (result[0] == NULL))
			papiAttributeListGetString(
					INT_JOB_HANDLE(job)->attributes, NULL,
					"job-printer-uri", &result);
	}
	return(result);
}

/**********************************************************************
 * papiJobGetId
 *********************************************************************/
int32_t papiJobGetId(papi_job_t job)
{
	if (job)
		return( INT_JOB_HANDLE(job)->id );
	else
		return(0);
}

/**********************************************************************
 * papiJobGetJobTicket
 *********************************************************************/
papi_job_ticket_t* papiJobGetJobTicket(papi_job_t job)
{
	if (job)
		return( INT_JOB_HANDLE(job)->job_ticket );
	else
		return(NULL);
}

/**********************************************************************
 * papiJobFree
 *********************************************************************/
void papiJobFree(papi_job_t job)
{
	papi_int_job_t* int_job = INT_JOB_HANDLE(job);

	if (int_job)
	{
		if (int_job->name)
			free(int_job->name);
		if (int_job->attributes)
			papiAttributeListFree(int_job->attributes);
		if (int_job->job_ticket)
		{
			if (int_job->job_ticket->ticket_data)
				free(int_job->job_ticket->ticket_data);
			if (int_job->job_ticket->file_name)
				free(int_job->job_ticket->file_name);
			free(int_job->job_ticket);
		}
		free(int_job);
	}
}

/**********************************************************************
 * papiJobListFree
 *********************************************************************/
void papiJobListFree(papi_job_t* jobs )
{
	int i;

	if (jobs)
	{
		for(i=0; jobs[i] != NULL; i++)
			papiJobFree(jobs[i]);
	}
}
