/**********************************************************************
 *
 * Open Standard Print API (PAPI) - Printer Functions
 *
 * Copyright (c) 2002 International Business Machines Corp.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "papi.h"
#include "papi_int.h"

/*---------------------------------------------------------------------
 * Static variables
 *-------------------------------------------------------------------*/

/**********************************************************************
 *  papi_do_base_printer_op
 *********************************************************************/
papi_status_t papi_do_base_printer_op(
							  papi_int_service_t*	handle,
						char* 				name,
						ipp_op_t			    op
						)
{
	papi_status_t rc;
	char resource[HTTP_MAX_URI];

	papi_init();
	papi_dbg_printf(">papi_do_base_printer_op(0x%x)\n", op);

	/* Set up the request */
	rc = papi_set_up_request(handle, name, PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papi_do_base_printer_op", handle, rc);

	/* Add operation-specific data to the request */
	ippAddString(handle->request, IPP_TAG_OPERATION, IPP_TAG_URI,	
				"printer-uri", NULL, handle->object_uri);

	handle->request->request.op.operation_id = op;
	handle->request->request.any.request_id = 1;
	
	/* Do the request */
	if ((op == IPP_PAUSE_PRINTER)||(op == IPP_RESUME_PRINTER)||
	    (op == IPP_PURGE_JOBS))
		strcpy(resource, "/admin/");
	else
		strcpy(resource, "/printers");
 	papi_do_request(handle, resource, NULL);

	rc = papi_get_result(INT_HANDLE(handle));
	PAPI_RETURN("papi_do_base_printer_op", handle, rc);
}

/*---------------------------------------------------------------------
 * Printer API
 *-------------------------------------------------------------------*/

/**********************************************************************
 * papiPrintersList
 *********************************************************************/
/* This function is not part of the IPP standard, so look for it in  */
/* spooler-specific modules (e.g. papi_cups.c).                      */

/**********************************************************************
 * papiPrinterQuery
 *********************************************************************/
papi_status_t papiPrinterQuery(
                papi_service_t     handle,
          char*              name,
          char*              requested_attrs[],
          papi_attribute_t** job_attrs,
                papi_printer_t*    printer )
{
	papi_status_t rc;
	char resource[HTTP_MAX_URI];
	static char* all_attrs[] = { "all" }; /* Requested attributes */
	papi_int_printer_t* int_printer = NULL;

	papi_init();
	papi_dbg_printf(">papiPrinterQuery\n");
	*printer = NULL;

	/* Set up the request */
	rc = papi_set_up_request(INT_HANDLE(handle), name, PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papiPrinterQuery", handle, rc);

	/* Add operation-specific data to the request */
	ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_URI,	
				"printer-uri", NULL, INT_HANDLE(handle)->object_uri);
	if ((requested_attrs == NULL) || (requested_attrs[0] == NULL))
	{
		/* get all attributes */
		papi_dbg_printf("add req for all attributes\n");
		ippAddStrings(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_NAME,
	                "requested-attributes",
					sizeof(all_attrs) / sizeof(all_attrs[0]),
					NULL, all_attrs);
	}
	else
	{
		/* get requested attributes */
		papi_dbg_printf("add req for selected attributes\n");
		ippAddStrings(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_NAME,
	                "requested-attributes",
					papi_get_array_count(requested_attrs),
					NULL, requested_attrs);
	}

	INT_HANDLE(handle)->request->request.op.operation_id = IPP_GET_PRINTER_ATTRIBUTES;
	INT_HANDLE(handle)->request->request.any.request_id = 1;
	
	/* Do the query */
 	papi_do_request(INT_HANDLE(handle), "/printers", NULL);
 	if (INT_HANDLE(handle)->response == NULL)
		PAPI_RETURN("papiPrinterQuery", handle, PAPI_BAD_REQUEST);

	rc = papi_get_result(INT_HANDLE(handle));

	/* Create the printer object to be returned */
	int_printer = malloc(sizeof(papi_int_printer_t));
	if (int_printer == NULL)
		PAPI_RETURN("papiPrinterQuery", handle, PAPI_BAD_REQUEST);
	int_printer->name = strdup(name);
	int_printer->attributes = NULL;
	*printer = (papi_printer_t)int_printer;
	
	if (rc == PAPI_OK)
	{
		/* Convert returned attributes to PAPI object/attributes */
		rc = papi_attr_cups2papi(INT_HANDLE(handle)->response->attrs,
                           &(int_printer->attributes));
	}

	if ((rc == PAPI_OK) && (job_attrs != NULL))
	{
		/* Indicate that we ignored job_attrs */
		rc = PAPI_OK_SUBST;
	}

	PAPI_RETURN("papiPrinterQuery", handle, rc);
}

/**********************************************************************
 * papiPrinterModify
 *********************************************************************/
papi_status_t papiPrinterModify(
                papi_service_t     handle,
          char*              name,
          papi_attribute_t** attrs,
                papi_printer_t*    printer )
{
	papi_status_t rc;
	char resource[HTTP_MAX_URI];
	papi_int_printer_t* int_printer = NULL;

	papi_init();
	papi_dbg_printf(">papiPrinterModify\n");
	*printer = NULL;

	/* Set up the request */
	rc = papi_set_up_request(INT_HANDLE(handle), name, PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papiPrinterModify", handle, rc);

	/* Add operation-specific data to the request */
	ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_URI,	
				"printer-uri", NULL, INT_HANDLE(handle)->object_uri);

	/* Add the set attributes */
	rc = papi_attr_papi2cups(attrs,
                             INT_HANDLE(handle)->request,
                             NULL,
                             IPP_TAG_PRINTER);
    if (rc != PAPI_OK)
		PAPI_RETURN("papiPrinterModify", handle, rc);

	INT_HANDLE(handle)->request->request.op.operation_id = IPP_SET_PRINTER_ATTRIBUTES;
	INT_HANDLE(handle)->request->request.any.request_id = 1;
	
	/* Do the modify */
 	papi_do_request(INT_HANDLE(handle), "/printers", NULL);
 	if (INT_HANDLE(handle)->response == NULL)
		PAPI_RETURN("papiPrinterModify", handle, PAPI_BAD_REQUEST);

	rc = papi_get_result(INT_HANDLE(handle));

	/* Create the printer object to be returned */
	int_printer = malloc(sizeof(papi_int_printer_t));
	if (int_printer == NULL)
		PAPI_RETURN("papiPrinterModify", handle, PAPI_BAD_REQUEST);
	int_printer->name = strdup(name);
	int_printer->attributes = NULL;
	*printer = (papi_printer_t)int_printer;
	
	if (rc == PAPI_OK)
	{
		/* Convert returned attributes to PAPI object/attributes */
		rc = papi_attr_cups2papi(INT_HANDLE(handle)->response->attrs,
                           &(int_printer->attributes));
	}

	PAPI_RETURN("papiPrinterModify", handle, rc);
}

/**********************************************************************
 * papiPrinterPause
 *********************************************************************/
papi_status_t papiPrinterPause(
                papi_service_t     handle,
          char*              name, char *message )
{
	return( papi_do_base_printer_op(INT_HANDLE(handle), name, IPP_PAUSE_PRINTER) );
}

/**********************************************************************
 * papiPrinterResume
 *********************************************************************/
papi_status_t papiPrinterResume(
                papi_service_t     handle,
          char*              name )
{
	return( papi_do_base_printer_op(INT_HANDLE(handle), name, IPP_RESUME_PRINTER) );
}

/**********************************************************************
 * papiPrinterPurgeJobs
 *********************************************************************/
papi_status_t papiPrinterPurgeJobs(
                papi_service_t     handle,
          char*              name,
                papi_job_t**       result )
{
	papi_status_t rc;

	rc = papi_do_base_printer_op(INT_HANDLE(handle), name, IPP_PURGE_JOBS);
	if ((rc == PAPI_OK) && (result != NULL))
		rc = PAPI_OK_SUBST;
	return(rc);
}

/**********************************************************************
 * papiPrinterListJobs
 *********************************************************************/
papi_status_t papiPrinterListJobs(
                papi_service_t     handle,
          char*              printer,
          char*              requested_attrs[],
          int                type_mask,
          int                max_num_jobs,
                papi_job_t**       jobs )
{
	papi_status_t rc;
	char resource[HTTP_MAX_URI];
	static char* all_attrs[] = { "all" }; /* Requested attributes */
	ipp_attribute_t* attr;
	ipp_attribute_t* jt_attr;
	int jt_found;
	papi_attribute_t* job_id_attr;
	int job_count = 0;
	int attr_count = 0;
	int i;
	papi_int_job_t** int_jobs;

	papi_init();
	papi_dbg_printf(">papiPrinterListJobs\n");
	*jobs = NULL;

	/* Set up the request */
	rc = papi_set_up_request(INT_HANDLE(handle), printer, PAPI_PRINTER);
	if (rc != PAPI_OK)
		PAPI_RETURN("papiPrinterListJobs", handle, rc);

	/* Add operation-specific data to the request */
	ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_URI,	
				"printer-uri", NULL, INT_HANDLE(handle)->object_uri);
	if ((requested_attrs == NULL) || (requested_attrs[0] == NULL))
	{
		/* get all attributes */
		papi_dbg_printf("add req for all attributes\n");
		ippAddStrings(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_NAME,
	                "requested-attributes",
					sizeof(all_attrs) / sizeof(all_attrs[0]),
					NULL, all_attrs);
	}
	else
	{
		/* get requested attributes */
		papi_dbg_printf("add req for selected attributes\n");
		ippAddStrings(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_NAME,
	                "requested-attributes",
					papi_get_array_count(requested_attrs),
					NULL, requested_attrs);
	}

	if ((type_mask & PAPI_LIST_JOBS_COMPLETED) && !(type_mask & PAPI_LIST_JOBS_NOT_COMPLETED))
		ippAddString(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,	
				"which-jobs", NULL, "completed");

	if ( !(type_mask & PAPI_LIST_JOBS_OTHERS) )
		ippAddBoolean(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, "my-jobs", 1);

	if (max_num_jobs)
		ippAddInteger(INT_HANDLE(handle)->request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,
						"limit", max_num_jobs);

	INT_HANDLE(handle)->request->request.op.operation_id = IPP_GET_JOBS;
	INT_HANDLE(handle)->request->request.any.request_id = 1;
	
	/* Do the query */
 	papi_do_request(INT_HANDLE(handle), "/printers", NULL);
	if (INT_HANDLE(handle)->response == NULL)
		PAPI_RETURN("papiPrinterListJobs", handle, PAPI_BAD_REQUEST);

	/* Count the number of job objects returned */
	for(attr=INT_HANDLE(handle)->response->attrs;	
		attr != NULL;
		attr=attr->next)
	{
		attr_count++;
		if ( (job_count == 0) ||
		     ((attr->name == NULL) && (attr->group_tag == IPP_TAG_ZERO)) )
			job_count++;
	}
	/* CUPS will return a couple generic attributes instead of "nothing" */
	if ((job_count == 1) && (attr_count < 3))
		job_count = 0;
	papi_dbg_printf("%d jobs returned\n", job_count);

	/* Allocate the jobs list */
	int_jobs = malloc(sizeof(papi_int_job_t*) * (job_count+1));
	if (int_jobs == NULL)
	{
		papi_dbg_printf("<papiPrinterListJobs returning PAPI_INTERNAL_ERROR at line %d\n", __LINE__);
		return(PAPI_INTERNAL_ERROR);
	}
	int_jobs[job_count] = NULL;
	*jobs = (papi_job_t*)int_jobs;

	/* For each job in the list, return its attributes */
	for(i=0, attr=INT_HANDLE(handle)->response->attrs; i < job_count; i++)
	{
		/* Allocate the job object to return */
		int_jobs[i] = malloc(sizeof(papi_int_job_t));
		if (int_jobs[i] == NULL)
			PAPI_RETURN("papiPrinterListJobs", handle, PAPI_INTERNAL_ERROR);

		/* Intialize the job information */
		int_jobs[i]->name = strdup(printer);
		int_jobs[i]->id = -1;
		int_jobs[i]->job_ticket = NULL;
		
		/* Convert the CUPS job attributes to PAPI attributes */
		rc = papi_attr_cups2papi(attr, &(int_jobs[i]->attributes));
		if (rc != PAPI_OK)
			PAPI_RETURN("papiPrinterListJobs", handle, rc);

		/* Set the returned job ID */
		job_id_attr = papiAttributeListFind(int_jobs[i]->attributes, "job-id");
		if (job_id_attr)
		{
			if (job_id_attr->values)
			{
				int_jobs[i]->id = job_id_attr->values[0]->integer;
				papi_dbg_printf("returning job id %d\n", int_jobs[i]->id);
			}
		}

		/* Set the returned job ticket */
		jt_found = 0;
		for(jt_attr=attr ;
			(jt_attr != NULL) &&
		     !((jt_attr->name == NULL) && (attr->group_tag == IPP_TAG_ZERO));
			jt_attr=jt_attr->next)
		{
			if ( (jt_attr->name != NULL) && !strcmp(jt_attr->name, "job-ticket") )
			{
				jt_found = 1;
				break;
			}
		}
		if ( !jt_found )
		{
			for(jt_attr=attr ;
				(jt_attr != NULL) &&
			     !((jt_attr->name == NULL) && (attr->group_tag == IPP_TAG_ZERO));
				jt_attr=jt_attr->next)
			{
				if ( (jt_attr->name != NULL) && !strcmp(jt_attr->name, "job-ticket-reference") )
				{
					jt_found = 2;
					break;
				}
			}
		}
		if (jt_found)
		{
			int_jobs[i]->job_ticket = malloc(sizeof(papi_job_ticket_t));
			papi_job_ticket_cups2jt(jt_attr, int_jobs[i]->job_ticket);
		}

		/* Access the start of the next job's attributes */
		for(attr=attr->next ;
			(attr != NULL) &&
		     !((attr->name == NULL) && (attr->group_tag == IPP_TAG_ZERO));
			attr=attr->next) ;
		if (attr != NULL)
			attr = attr->next;
	}

	PAPI_RETURN("papiPrinterListJobs", handle, rc);
}

/**********************************************************************
 * papiPrinterGetAttributeList
 *********************************************************************/
papi_attribute_t** papiPrinterGetAttributeList(papi_printer_t printer)
{
	if (printer)
		return( INT_PRTR_HANDLE(printer)->attributes );
	else
		return(NULL);
}

/**********************************************************************
 * papiPrinterFree
 *********************************************************************/
void papiPrinterFree(papi_printer_t printer)
{
	papi_int_printer_t* int_printer = INT_PRTR_HANDLE(printer);

	if (int_printer)
	{
		if (int_printer->attributes)
			papiAttributeListFree(int_printer->attributes);
		if (int_printer->name)
			free(int_printer->name);
		free(int_printer);
	}
}

/**********************************************************************
 * papiPrinterListFree
 *********************************************************************/
void papiPrinterListFree(papi_printer_t* printers)
{
	int i;

	if (printers)
	{
		for(i=0; printers[i] != NULL; i++)
			papiPrinterFree(printers[i]);
	}
}
